# coding: utf-8
from __future__ import annotations


import functools
from typing import Iterable

from silx.gui import qt

from tomwer.core.process.reconstruction.nabu.utils import _NabuStages
from tomwer.gui import icons
from tomwer.utils import docstring

from ...utils.flow import FlowCanvas, FlowDirection


class NabuFlowControl(qt.QWidget):
    """
    Widget which is composed of three `FlowCanvas`: one for pre processing
    (where dockers are not movable), one for processing (where dockers are
    movable) and one for post-processing (where dockers are not movable)
    """

    sigConfigurationChanged = qt.Signal(tuple)
    """signal emitted when the `focus` process on the flow is changed.
    tuple is (stage, option)
    """
    sigResetConfiguration = qt.Signal()
    """Signal emitted when the configuration should be reset"""

    def __init__(self, parent, direction):
        qt.QWidget.__init__(self, parent=parent)

        self._direction = FlowDirection(direction)

        if self._direction is FlowDirection.VERTICAL:
            self.setLayout(qt.QVBoxLayout())
        else:
            self.setLayout(qt.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        # initialization
        self._iniProcessWidgets = FlowCanvas(
            direction=direction, parent=self, show_lock_state=False
        )

        # preprocessing flow
        self._preProcessWidgets = FlowCanvas(
            direction=direction, parent=self, show_lock_state=False
        )

        # phase processing flow
        self._phaseProcessWidgets = FlowCanvas(
            direction=direction, parent=self, show_lock_state=False
        )

        # processing flow
        self._processWidgets = FlowCanvas(
            direction=direction, parent=self, show_lock_state=False
        )

        # postprocessing flow
        self._postProcessWidgets = FlowCanvas(
            direction=direction, parent=self, show_lock_state=False
        )

        for widget in (
            self._iniProcessWidgets,
            self._preProcessWidgets,
            self._phaseProcessWidgets,
            self._processWidgets,
            self._postProcessWidgets,
        ):
            self.layout().addWidget(widget)

        spacer = qt.QWidget(parent=self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer)

        # connect signal / slot
        self._iniProcessWidgets.sigWidgetActivated.connect(
            functools.partial(self._updateActiveProcess, _NabuStages.INI)
        )
        self._iniProcessWidgets.sigWidgetDeactivated.connect(self.sigResetConfiguration)
        self._preProcessWidgets.sigWidgetActivated.connect(
            functools.partial(self._updateActiveProcess, _NabuStages.PRE)
        )
        self._preProcessWidgets.sigWidgetDeactivated.connect(self.sigResetConfiguration)
        self._phaseProcessWidgets.sigWidgetActivated.connect(
            functools.partial(self._updateActiveProcess, _NabuStages.PHASE)
        )
        self._phaseProcessWidgets.sigWidgetDeactivated.connect(
            self.sigResetConfiguration
        )
        self._processWidgets.sigWidgetActivated.connect(
            functools.partial(self._updateActiveProcess, _NabuStages.PROC)
        )
        self._processWidgets.sigWidgetDeactivated.connect(self.sigResetConfiguration)
        self._postProcessWidgets.sigWidgetActivated.connect(
            functools.partial(self._updateActiveProcess, _NabuStages.POST)
        )
        self._postProcessWidgets.sigWidgetDeactivated.connect(
            self.sigResetConfiguration
        )

        # set up
        self.setIniprocVisible(False)
        self.setPreProcVisible(True)
        self.setPhaseVisible(True)
        self.setProcVisible(True)
        self.setPostVisible(True)

    def _updateActiveProcess(self, stage):
        stage = _NabuStages(stage)

        if stage is _NabuStages.INI:
            activeWidget = self._iniProcessWidgets
        elif stage is _NabuStages.PRE:
            activeWidget = self._preProcessWidgets
        elif stage is _NabuStages.PHASE:
            activeWidget = self._phaseProcessWidgets
        elif stage is _NabuStages.PROC:
            activeWidget = self._processWidgets
        elif stage is _NabuStages.POST:
            activeWidget = self._postProcessWidgets
        else:
            raise ValueError("nabu stage not recognized")

        for widget in (
            self._iniProcessWidgets,
            self._preProcessWidgets,
            self._phaseProcessWidgets,
            self._processWidgets,
            self._postProcessWidgets,
        ):
            if widget != activeWidget:
                widget.clearActiveProcess()
        self.sigConfigurationChanged.emit(self.getProcessFocused())

    def setIniprocVisible(self, visible):
        """
        change visibility of the widgets relative to initialization process

        :param visible:
        """
        self._iniProcessWidgets.setVisible(visible)

    def setPreProcVisible(self, visible):
        """
        change visibility of the widgets relative to pre processing process

        :param visible:
        """
        self._preProcessWidgets.setVisible(visible)

    def setPhaseVisible(self, visible: bool):
        """
        change visibility of the widgets relative to pre processing process
        """
        self._phaseProcessWidgets.setVisible(visible)

    def setProcVisible(self, visible: bool):
        """
        change visibility of the widgets relative to processing process
        """
        self._processWidgets.setVisible(visible)

    def setPostVisible(self, visible: bool):
        """
        change visibility of the widgets relative to post processing process
        """
        self._postProcessWidgets.setVisible(visible)

    def setIniProcessing(self, processes: Iterable, icons: Iterable):
        """
        Defines the processing to execute for initialization

        :param processes: processes to display
        :param icons: icons associated to the processes if any
        """
        draggability = [False] * len(processes)
        self._iniProcessWidgets.setProcesses(
            processes=processes, icons=icons, draggability=draggability
        )
        self.setIniprocVisible(len(processes) != 0)

    def setPreProcessing(self, processes: Iterable, icons: Iterable):
        """

        :param processes: processes to display
        :param icons: icons associated to the processes if any
        """
        draggability = [False] * len(processes)
        self._preProcessWidgets.setProcesses(
            processes=processes, icons=icons, draggability=draggability
        )
        self.setPreProcVisible(len(processes) != 0)

    def setPhaseProcessing(self, processes: Iterable, icons: Iterable):
        """

        :param processes: processes to display
        :param icons: icons associated to the processes if any
        """
        draggability = [False] * len(processes)
        self._phaseProcessWidgets.setProcesses(
            processes=processes, icons=icons, draggability=draggability
        )
        self.setPhaseVisible(len(processes) != 0)

    def setProcessing(self, processes: Iterable, icons: Iterable):
        """

        :param processes: processes to display
        :param icons: icons associated to the processes if any
        """
        draggability = [False] * len(processes)
        self._processWidgets.setProcesses(
            processes=processes, icons=icons, draggability=draggability
        )
        self.setProcVisible(len(processes) != 0)

    def setPostProcessing(self, processes: Iterable, icons: Iterable):
        """

        :param processes: processes to display
        :param icons: icons associated to the processes if any
        """
        draggability = [False] * len(processes)
        self._postProcessWidgets.setProcesses(
            processes=processes, icons=icons, draggability=draggability
        )
        self.setPostVisible(len(processes) != 0)

    def getIniProcessingFlow(self) -> list:
        """

        :return: list of initialization processes
        """
        return self._iniProcessWidgets.getFlow()

    def getPreProcessingFlow(self):
        """

        :return: list of preprocessing
        """
        return self._preProcessWidgets.getFlow()

    def getProcessingFlow(self):
        """

        :return: list of processing
        """
        return self._processWidgets.getFlow()

    def getPostProcessingFlow(self):
        """

        :return: list of postprocessing
        """
        return self._postProcessWidgets.getFlow()

    def getFlow(self) -> dict:
        """
        Returns the concatenation of pre-processing, processing, and post-processing flows.

        :return: Dictionary where the keys are Nabu stages, and the values are lists of flows for each stage.
        :rtype: dict

        This method fetches the processing flows for all stages and returns them
        as a dictionary, allowing easy access to the current configuration.
        """
        return {
            _NabuStages.INI: self._iniProcessWidget.getFlow(),
            _NabuStages.PRE: self._preProcessWidgets.getFlow(),
            _NabuStages.PHASE: self._phaseProcessWidget.getFlow(),
            _NabuStages.PROC: self._processWidgets.getFlow(),
            _NabuStages.POST: self._processWidgets.getFlow(),
        }

    def getProcessFocused(self) -> tuple:
        """

        :return: the process under focus
        """
        stages = (
            _NabuStages.INI,
            _NabuStages.PRE,
            _NabuStages.PHASE,
            _NabuStages.PROC,
            _NabuStages.POST,
        )
        widgets = (
            self._iniProcessWidgets,
            self._preProcessWidgets,
            self._phaseProcessWidgets,
            self._processWidgets,
            self._postProcessWidgets,
        )
        for stage, widget in zip(stages, widgets):
            process_id = widget.getProcessFocused()
            if process_id is not None:
                return stage, process_id
        return None, None

    def _removeActiveProcess(self):
        for widget in (
            self._iniProcessWidgets,
            self._preProcessWidgets,
            self._phaseProcessWidgets,
            self._processWidgets,
            self._postProcessWidgets,
        ):
            if widget.hasProcessFocus():
                widget._removeActiveProcess()

        self.sigConfigurationChanged.emit((None, None))


class NabuFlowArea(qt.QWidget):
    """Define the flow dedicated to nabu"""

    sigConfigurationChanged = qt.Signal(tuple)
    """signal emitted when the `focus` process on the flow is changed.
    tuple is (stage, option)"""
    sigResetConfiguration = qt.Signal()
    """Signal emitted when the configuration should be reset"""

    def __init__(self, parent, direction):
        qt.QWidget.__init__(self, parent=parent)
        direction = FlowDirection(direction)

        if direction is FlowDirection.VERTICAL:
            self.setLayout(qt.QHBoxLayout(self))
        else:
            self.setLayout(qt.QVBoxLayout(self))

        self._control = NabuFlowControl(parent=self, direction=direction)
        self._control.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)
        self._addRmWidget = _AddRemoveProcessWidget(parent=self, direction=direction)
        self._addRmWidget.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)
        self._control.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)

        if direction is FlowDirection.VERTICAL:
            # if vertical illustration left sided
            self.layout().addWidget(self._control)
            self.layout().addWidget(self._addRmWidget)
        else:
            # if horizontal illustration bottom sided
            self.layout().addWidget(self._control)
            self.layout().addWidget(self._addRmWidget)

        # set up
        # for now hide the add and remove option
        self._addRmWidget.setVisible(False)

        # connect signal / slot
        self._addRmWidget.sigRmProcess.connect(self._control._removeActiveProcess)
        self._control.sigConfigurationChanged.connect(self.sigConfigurationChanged)
        self._control.sigResetConfiguration.connect(self.sigResetConfiguration)

    @docstring(NabuFlowControl)
    def setIniProcessing(self, processes, icons):
        return self._control.setIniProcessing(processes=processes, icons=icons)

    @docstring(NabuFlowControl)
    def setPreProcessing(self, processes, icons):
        return self._control.setPreProcessing(processes=processes, icons=icons)

    @docstring(NabuFlowControl)
    def setPhaseProcessing(self, processes, icons):
        return self._control.setPhaseProcessing(processes=processes, icons=icons)

    @docstring(NabuFlowControl)
    def setProcessing(self, processes, icons):
        return self._control.setProcessing(processes=processes, icons=icons)

    @docstring(NabuFlowControl)
    def setPostProcessing(self, processes, icons):
        return self._control.setPostProcessing(processes=processes, icons=icons)

    @docstring(NabuFlowControl)
    def getFlow(self):
        return self._control.getFlow()

    @docstring(NabuFlowControl)
    def getProcessFocused(self):
        return self._control.getProcessFocused()


class _AddRemoveProcessWidget(qt.QWidget):
    sigAddProcess = qt.Signal()
    """Signal emitted when the user request a process to be added"""

    sigRmProcess = qt.Signal()
    """Signal emitted when the user request a process to be removed"""

    def __init__(self, parent, direction):
        qt.QWidget.__init__(self, parent=parent)
        if direction is FlowDirection.VERTICAL:
            self.setLayout(qt.QVBoxLayout())
            spacer_size_policy = (qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        else:
            self.setLayout(qt.QHBoxLayout())
            spacer_size_policy = (qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        add_icon = icons.getQIcon("add")
        rm_icon = icons.getQIcon("rm")

        self._addButton = qt.QPushButton(add_icon, "", self)
        self._addButton.setToolTip("add process")
        self._rmButton = qt.QPushButton(rm_icon, "", self)
        self._rmButton.setToolTip("remove process")
        spacer = qt.QWidget(parent=self)
        spacer.setSizePolicy(*spacer_size_policy)
        self.layout().addWidget(spacer)
        self.layout().addWidget(self._addButton)
        self.layout().addWidget(self._rmButton)

        # connect signal / slot
        self._addButton.pressed.connect(self._addProcess)
        self._rmButton.pressed.connect(self.sigRmProcess)

    def _addProcess(self):
        # should be able to add a process from ini, post, processing and pre

        raise NotImplementedError(" not implemented, to be defined")



from tomwer.gui.reconstruction.nabu.nabuconfig.preprocessing import (
    SinoRingsOptions,
    TiltCorrection,
    RingCorrectionMethod,
)
from tomwer.tests.conftest import qtapp  # noqa F401


def test_tilt_correction_widget(qtapp):  # noqa F811
    """
    Test of TiltCorrection
    """
    widget = TiltCorrection(text="")
    assert not widget._tiltManualRB.isChecked()
    assert widget._autoManualRB.isChecked()
    assert widget.getTiltCorrection() == ("1d-correlation", "")
    widget.setTiltCorrection(0.0)
    assert widget.getTiltCorrection() == (0.0, "")
    assert widget._tiltManualRB.isChecked()
    assert not widget._autoManualRB.isChecked()
    widget.setTiltCorrection("fft-polar", auto_tilt_options="low_pass=10")
    assert widget.getTiltCorrection() == ("fft-polar", "low_pass=10")
    assert not widget._tiltManualRB.isChecked()
    assert widget._autoManualRB.isChecked()


def test_sino_rings_options(qtapp):  # noqa F811
    """
    Test of SinoRingsOptions
    """
    widget = SinoRingsOptions()
    # test munch options
    assert widget.getOptions() == {
        "sigma": 1.0,
        "levels": 10,
        "padding": False,
    }

    widget.setOptions(
        {
            "sigma": 3.25,
            "levels": 564,
            "padding": True,
        }
    )
    assert widget.getOptions() == {
        "sigma": 3.25,
        "levels": 564,
        "padding": True,
    }
    # test mean division
    widget.setMethod(method=RingCorrectionMethod.MEAN_DIVISION)
    assert widget.getOptions() == {
        "filter_cutoff": (0, 30),
    }

    # test VO deringer
    widget.setMethod(method=RingCorrectionMethod.VO.value)
    assert widget.getOptions() == {
        "dim": 1,
        "la_size": 51,
        "sm_size": 21,
        "snr": 3.0,
    }
    new_vo_options = {
        "dim": 2,
        "la_size": 42,
        "sm_size": 22,
        "snr": 6.3,
    }
    widget.setOptions(new_vo_options)
    assert widget.getOptions() == new_vo_options

    # test mean division
    new_mean_division_options = {
        "filter_cutoff": (10, 23),
    }

    widget.setMethod(method=RingCorrectionMethod.MEAN_DIVISION)
    widget.setOptions(new_mean_division_options)
    assert widget.getOptions() == new_mean_division_options



"""
Some utils GUI associated to illustrations
"""

from __future__ import annotations
from silx.io.url import DataUrl


import logging
import os
import subprocess
import threading
import platform

_logger = logging.getLogger(__name__)


__has_image_j = None
__image_j_from = None

_IMAGE_J_BIN_PATH = "/sware/pub/ImageJ/ImageJ"


def has_imagej() -> bool:
    """Return if imagej command is accessible from the computer"""
    global __has_image_j

    def has_bin() -> bool:
        global __image_j_from
        if platform.machine() in ("x86_64", "AMD64") and os.path.exists(
            _IMAGE_J_BIN_PATH
        ):
            __image_j_from = "binary"
            return True
        return False

    def has_CLI() -> bool:
        global __image_j_from
        try:
            # use help because there is no information regarding version
            subprocess.call(
                ["imagej", "-h"], stdout=subprocess.PIPE
            )  # nosec B603, B607
        except Exception:
            return False
        else:
            __image_j_from = "command"
            return True

    def has_module() -> bool:
        """Check if the module exists. At the moment this is the simplest way to access as utils like is-loaded seems to fail."""
        global __image_j_from
        if os.path.exists("/cvmfs/hpc.esrf.fr/software/packages/linux/x86_64/imagej"):
            __image_j_from = "command"
            return True
        else:
            return False

    if __has_image_j is None:
        __has_image_j = has_bin() or has_CLI() or has_module()

    return __has_image_j


def open_url_with_image_j(url: DataUrl) -> threading.Thread:
    """open the url in an imagej subprocess within a thread.
    It is up to the caller to handle thread life

    :param url: url we want to open in imagej
    """
    global __image_j_from  # noqa: F824
    if not has_imagej():
        raise OSError("ImageJ is not installed")
    thread = ImageJthread(url=url, image_j_from=__image_j_from)
    thread.start()
    return thread


class ImageJthread(threading.Thread):
    def __init__(self, url, image_j_from, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._image_j_from = image_j_from
        self._url = url
        self._process = None

    def run(self):
        if self._image_j_from == "command":
            # for now we only manage the simple case of an edf file
            try:
                self._process = subprocess.Popen(
                    ["imagej", "-o", self._url.file_path()],
                )
            except Exception as e:
                _logger.warning(f"Fail to open {self._url}. Reason is {e}")
            else:
                self._process.communicate()

        elif self._image_j_from == "binary":
            try:
                self._process = subprocess.Popen(
                    [_IMAGE_J_BIN_PATH, self._url.file_path()],
                )
            except Exception as e:
                _logger.warning(f"Fail to open {self._url}. Reason is {e}")
            else:
                self._process.communicate()

    def quit(self):
        if self._process is not None:
            self._process.kill()
            self._process.wait()



from tomoscan.esrf.identifier.edfidentifier import EDFVolumeIdentifier
from tomoscan.esrf.identifier.hdf5Identifier import (
    HDF5VolumeIdentifier,
)
from tomoscan.esrf.identifier.jp2kidentifier import JP2KVolumeIdentifier
from tomoscan.esrf.identifier.rawidentifier import RawVolumeIdentifier
from tomoscan.esrf.identifier.tiffidentifier import (
    MultiTiffVolumeIdentifier,
    TIFFVolumeIdentifier,
)

from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.volume.volumefactory import VolumeFactory
from tomoscan.esrf.volume.utils import guess_volumes

from tomwer.core.volume.edfvolume import EDFVolume
from tomwer.core.volume.hdf5volume import HDF5Volume
from tomwer.core.volume.jp2kvolume import JP2KVolume
from tomwer.core.volume.rawvolume import RawVolume
from tomwer.core.volume.tiffvolume import MultiTIFFVolume, TIFFVolume


DEFAULT_SCHEME_TO_VOL = {
    EDFVolumeIdentifier.scheme: EDFVolume,
    HDF5VolumeIdentifier.scheme: HDF5Volume,
    TIFFVolumeIdentifier.scheme: TIFFVolume,
    MultiTiffVolumeIdentifier.scheme: MultiTIFFVolume,
    JP2KVolumeIdentifier.scheme: JP2KVolume,
    RawVolumeIdentifier.scheme: RawVolume,
}


def get_tomo_objs_instances(tomo_objs: tuple) -> tuple:
    """
    return instances of TomoObj from a tuple of string (either from identifiers directly or from file path parsing)
    """
    instances = []
    has_scans = False
    has_vols = False
    for tomo_obj in tomo_objs:

        def get_scans():
            try:
                return (ScanFactory.create_tomo_object_from_identifier(tomo_obj),)
            except Exception:
                try:
                    return ScanFactory.create_scan_objects(tomo_obj)
                except Exception:
                    return tuple()

        def get_volumes():
            try:
                return (VolumeFactory.create_tomo_object_from_identifier(tomo_obj),)
            except Exception:
                try:
                    return guess_volumes(
                        tomo_obj,
                        scheme_to_vol=DEFAULT_SCHEME_TO_VOL,
                        filter_histograms=True,
                    )
                except Exception:
                    return tuple()

        scans_found = get_scans()
        if len(scans_found) == 0:
            # we start by scan because they have a 'definition' as NXtomo available.
            # otherwise we might get some entries defined as scan AND volume...
            # for stitching we expect users to ask of for one or the other.
            volumes = get_volumes() or ()
            if len(volumes) > 0:
                has_vols = True
                instances.extend(volumes)
        else:
            has_scans = len(scans_found) > 0
            instances.extend(scans_found)

    return tuple(instances), (has_scans, has_vols)



from __future__ import annotations

import psutil
import logging
from psutil import process_iter
import getpass


_logger = logging.getLogger(__name__)


def send_signal_to_local_rpc_servers(signal, port: int, extended_find: bool = True):
    """
    :param signal: signal to be emit
    :param port: port to check
    :param extended_find: if True then will try to find a process that occupy the port
        even if this is not a process launched by the user and launched by tomwer.
    """

    found = False
    for proc in process_iter():
        # try to find a process we can handle
        if proc.username() == getpass.getuser():
            try:
                for conns in proc.connections():
                    # make sure we will kill the correct process
                    if conns.laddr.port == port and proc.name() in (
                        "tomwer",
                        "orange-canvas",
                    ):
                        _logger.warning(f"send {signal} signal to pid {proc.pid}")
                        proc.send_signal(signal)
                        found = True
                        return
            except (PermissionError, psutil.AccessDenied):
                pass
    if not extended_find:
        return
    # if process not found try to find one to inform the user
    if not found:
        for proc in process_iter():
            try:
                for conns in proc.connections():
                    # make sure we will kill the correct process
                    if conns.laddr.port == port:
                        _logger.warning(
                            f"process pid: {proc.pid} - {proc.name()} seems to be one occupying port {port}"
                        )
                        return
            except (PermissionError, psutil.AccessDenied):
                pass



# coding: utf-8

from __future__ import annotations


import logging
import os
import pathlib

from tomoscan.series import Series

from nxtomomill import converter as nxtomomill_converter
from nxtomomill.models.h52nx import H52nxModel
from nxtomomill.models.edf2nx import EDF2nxModel

from nxtomomill.converter.hdf5.utils import (
    get_default_output_file,
)

from tomwer.io.utils.raw_and_processed_data import to_raw_data_path
from tomwer.core.process.output import ProcessDataOutputDirMode
from tomwer.core.process.task import TaskWithProgress
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.utils.scanutils import format_output_location

_logger = logging.getLogger(__name__)


class H5ToNxProcess(
    TaskWithProgress,
    input_names=("h5_to_nx_configuration",),
    optional_input_names=(
        "progress",
        "bliss_scan",
        "serialize_output_data",
    ),
    output_names=("data", "series"),
):
    """
    Task to convert from a bliss dataset to a nexus compliant dataset
    """

    @staticmethod
    def deduce_output_file_path(master_file_name, scan, output_dir):
        master_file_name = os.path.abspath(master_file_name)
        # step 1: get output dir
        try:
            output_dir = ProcessDataOutputDirMode.from_value(output_dir)
        except ValueError:
            # case path provided directly
            output_folder = format_output_location(output_dir, scan=scan)
        else:
            if output_dir is ProcessDataOutputDirMode.OTHER:
                raise ValueError(
                    f"When output dir mode is {ProcessDataOutputDirMode.OTHER} we expect to received the output dir directly"
                )
            if output_dir is ProcessDataOutputDirMode.PROCESSED_DATA_FOLDER:
                path = pathlib.Path(
                    get_default_output_file(input_file=master_file_name)
                )
                output_folder = str(path.parent)
            elif output_dir is ProcessDataOutputDirMode.IN_SCAN_FOLDER:
                output_folder = os.path.dirname(master_file_name)
            else:
                raise RuntimeError(f"output dir {output_dir} not handled")

        # deduce output file from file basename and output directory
        file_basename = os.path.basename(master_file_name)
        file_name = os.path.splitext(file_basename)[0]
        file_name = file_name.replace(".", "_")
        file_name = file_name.replace(":", "_")
        output_file_name = file_name + ".nx"
        return os.path.join(output_folder, output_file_name)

    def run(self):
        config = self.inputs.h5_to_nx_configuration
        if isinstance(config, dict):
            config = H52nxModel.from_dict(config)
        elif not isinstance(config, H52nxModel):
            raise TypeError(
                f"h5_to_nx_configuration should be a dict or an instance of {H52nxModel}"
            )
        config.bam_single_file = True
        config.no_master_file = True
        try:
            convs = nxtomomill_converter.from_h5_to_nx(
                configuration=config, progress=self.progress
            )
        except Exception as e:
            _logger.error(e)
            return

        if len(convs) == 0:
            return

        series = []
        for conv in convs:
            conv_file, conv_entry = conv
            scan_converted = NXtomoScan(scan=conv_file, entry=conv_entry)
            _logger.processSucceed(
                f"{config.input_file} {config.entries} has been translated to {scan_converted}"
            )
            if self.get_input_value("serialize_output_data", True):
                data = scan_converted.to_dict()
            else:
                data = scan_converted
            series.append(data)
        self.outputs.series = Series(
            name=f"series created from {config.input_file}", iterable=series
        )
        self.outputs.data = series[-1] if len(series) > 0 else None


class EDFToNxProcess(
    TaskWithProgress,
    input_names=("edf_to_nx_configuration",),
    optional_input_names=(
        "progress",
        "edf_scan",
        "serialize_output_data",
    ),
    output_names=("data",),
):
    """
    Task calling edf2nx in order to insure conversion from .edf to .nx (create one NXtomo to be used elsewhere)
    """

    def run(self):
        config = self.inputs.edf_to_nx_configuration
        if isinstance(config, dict):
            config = EDF2nxModel.from_dict(config)
        elif not isinstance(config, EDF2nxModel):
            raise TypeError(
                f"edf_to_nx_configuration should be a dict or an instance of {EDF2nxModel}"
            )
        os.makedirs(os.path.dirname(config.output_file), exist_ok=True)
        file_path, entry = nxtomomill_converter.from_edf_to_nx(
            configuration=config, progress=self.progress
        )
        scan = NXtomoScan(entry=entry, scan=file_path)
        if self.get_input_value("serialize_output_data", True):
            self.outputs.data = scan.to_dict()
        else:
            self.outputs.data = scan

    @staticmethod
    def deduce_output_file_path(
        folder_path, output_dir: ProcessDataOutputDirMode | str, scan
    ):
        try:
            output_dir = ProcessDataOutputDirMode.from_value(output_dir)
        except Exception:
            pass

        if output_dir is ProcessDataOutputDirMode.OTHER:
            raise ValueError("if mode is other, we expect 'output_dir' to be the path")
        elif output_dir in (None, ProcessDataOutputDirMode.IN_SCAN_FOLDER):
            output_folder = os.path.dirname(folder_path)
        elif output_dir is ProcessDataOutputDirMode.PROCESSED_DATA_FOLDER:
            path = pathlib.Path(get_default_output_file(folder_path))
            output_folder = str(path.parent)
        else:
            # else we expect people to provide output location
            output_folder = format_output_location(output_dir, scan=scan)
        return os.path.join(output_folder, os.path.basename(folder_path) + ".nx")


def get_default_raw_data_output_file(
    input_file: str, output_file_extension: str = ".nx"
) -> str:
    """
    Policy: look for any 'RAW_DATA' in file directory. If find any (before any 'PROCESSED_DATA' directory) replace it "RAW_DATA".
    Then replace input_file by the expected file_extension and make sure the output file is different than the input file. Else append _nxtomo to it.

    :param input_file: file to be converted from bliss to NXtomo
    :param output_file_extension:
    :return: default output file according to policy
    """
    if isinstance(input_file, pathlib.Path):
        input_file = str(input_file)
    if not isinstance(input_file, str):
        raise TypeError(
            f"input_file is expected to be an instance of str. {type(input_file)} provided"
        )
    if not isinstance(output_file_extension, str):
        raise TypeError("output_file_extension is expected to be a str")
    if not output_file_extension.startswith("."):
        output_file_extension = "." + output_file_extension

    input_file = os.path.abspath(input_file)
    input_file_no_ext, _ = os.path.splitext(input_file)

    output_path = to_raw_data_path(input_file_no_ext)
    output_file = output_path + output_file_extension
    if output_file == input_file:
        # to be safer if the default output file is the same as the input file (if the input file has a .nx extension and not in any 'RAw_DATA' directory)
        return output_path + "_nxtomo" + output_file_extension
    else:
        return output_file



from __future__ import annotations

import os
from enum import Enum
import warnings


class ProcessDataOutputDirMode(Enum):
    IN_SCAN_FOLDER = "same folder as scan"
    PROCESSED_DATA_FOLDER = "PROCESSED_DATA folder"
    OTHER = "other"

    @classmethod
    def from_value(cls, value):
        # ensure backward compatibility
        if value == "near input":
            return cls.IN_SCAN_FOLDER
        elif value == "processed data dir":
            return cls.PROCESSED_DATA_FOLDER
        elif value == "raw data dir":
            # Deprecation logic
            warnings.warn(
                "'RAW_DATA output' is deprecated since writing to 'RAW_DATA' folder is forbidden."
                "Redirecting to 'PROCESSED_DATA_FOLDER'.",
                DeprecationWarning,
                stacklevel=2,
            )
            return cls.PROCESSED_DATA_FOLDER
        return ProcessDataOutputDirMode(value)


class NabuOutputFileFormat(Enum):
    TIFF = "tiff"
    TIFF_3D = "tiff3d"
    HDF5 = "hdf5"
    JP2K = "jp2"
    EDF = "edf"
    RAW = "vol"

    @classmethod
    def from_value(cls, value):
        if isinstance(value, str):
            value = value.lstrip(".")
        return NabuOutputFileFormat(value)


def get_file_format(file_str):
    extension = os.path.splitext(file_str.lower())[-1]
    extension = extension.lstrip(".")
    if extension in ("tiff", "tif"):
        return NabuOutputFileFormat.TIFF
    elif extension in ("hdf5", "hdf", "h5"):
        return NabuOutputFileFormat.HDF5
    elif extension in ("jp2", "jp2k", "jpg2k"):
        return NabuOutputFileFormat.JP2K
    elif extension in ("edf",):
        return NabuOutputFileFormat.EDF
    elif extension in ("vol", "raw"):
        return NabuOutputFileFormat.RAW
    else:
        raise ValueError(f"Unrecognized file extension {extension} from {file_str}")



from __future__ import annotations

import functools
import logging

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output

from processview.core import helpers as pv_helpers
from processview.core.manager import DatasetState, ProcessManager
from silx.gui import qt

import tomwer.core.process.reconstruction.saaxis.saaxis
from orangecontrib.tomwer.orange.managedprocess import SuperviseOW
from orangecontrib.tomwer.orange.settings import CallbackSettingsHandler
from tomwer.core import settings
from tomwer.core.utils.lbsram import is_low_on_memory
from tomwer.core.cluster import SlurmClusterConfiguration
from tomwer.core.process.reconstruction.axis import AxisTask
from tomwer.core.process.reconstruction.params_cache import (
    save_reconstruction_parameters_to_cache,
)
from tomwer.core.scan.scanbase import TomwerScanBase, _TomwerBaseDock
from tomwer.gui.reconstruction.saaxis.saaxis import SAAxisWindow as _SAAxisWindow
from tomwer.synctools.axis import QAxisRP
from tomwer.synctools.saaxis import QSAAxisParams
from tomwer.synctools.stacks.reconstruction.saaxis import SAAxisProcessStack

from ..utils import WidgetLongProcessing

_logger = logging.getLogger(__name__)


class SAAxisWindow(_SAAxisWindow):
    sigResultsToBeShow = qt.Signal()
    """signal emit when some results are ready to be display"""

    def __init__(self, Outputs, parent=None, process_id=None):
        _SAAxisWindow.__init__(self, parent=parent)
        self.Outputs = Outputs
        self._saaxis_params = QSAAxisParams()
        self._processing_stack = SAAxisProcessStack(
            saaxis_params=self._saaxis_params, process_id=process_id
        )
        if process_id is not None:
            assert self._processing_stack.process_id == process_id

        # connect signal / slot
        self.sigValidated.connect(self.validateCurrentScan)
        self._clusterConfig = None

    def setClusterConfig(self, cluster_config: dict):
        if not isinstance(
            cluster_config, (dict, type(None), SlurmClusterConfiguration)
        ):
            raise TypeError(
                f"cluster config is expected to be None, dict, {SlurmClusterConfiguration} not {type(cluster_config)}"
            )
        self._clusterConfig = cluster_config

    def _computeEstimatedCor(self) -> float | None:
        """callback when calculation of a estimated cor is requested"""
        # TODO: check but should not be needed anymore
        scan = self.getScan()
        if scan is None:
            text = (
                "No scan is set on the widget currently. No automatic center "
                "of rotation to be computed"
            )
            _logger.warning(text)
            msg = qt.QMessageBox(self)
            msg.setIcon(qt.QMessageBox.Warning)
            msg.setText(text)
            msg.show()
            return
        else:
            axis_params_info = self.getQAxisRP().get_simple_str()

            text = f"start automatic cor for {scan} with {axis_params_info}"
            _logger.inform(text)
            cor_estimation_process = AxisTask(
                inputs={
                    "axis_params": self.getQAxisRP(),
                    "data": scan,
                    "wait": True,
                    "serialize_output_data": False,
                },
                process_id=-1,
            )
            try:
                cor_estimation_process.run()
            except Exception as e:
                text = f"Unable to run automatic cor calculation. Reason is {e}"
                _logger.error(text)
                msg = qt.QMessageBox(self)
                msg.setIcon(qt.QMessageBox.Critical)
                msg.setText(text)
                msg.show()
                return None
            else:
                cor = scan.axis_params.relative_cor_value
                text = f"automatic cor computed for {scan}: {cor} ({axis_params_info})"
                _logger.inform(text)
                self.setEstimatedCorPosition(value=cor)
                self.getAutomaticCorWindow().hide()
                return cor

    def _launchReconstructions(self):
        """callback when we want to launch the reconstruction of the
        slice for n cor value"""
        scan = self.getScan()
        if scan is None:
            return
        # step1: if isAutoFocus: validate automatically the scan
        # step2: update the interface if the current scan is the one displayed
        # else skip it
        callback = functools.partial(
            self._mightUpdateResult, scan, self.isAutoFocusLock()
        )
        self._processing_stack.add(
            data=scan, configuration=self.getConfiguration(), callback=callback
        )

    def _validate(self):
        # TODO: check but should not be needed anymore
        pass

    def _mightUpdateResult(self, scan: TomwerScanBase, validate: bool):
        if not isinstance(scan, TomwerScanBase):
            raise TypeError("scan is expected to be an instance of TomwerScanBase")
        if not isinstance(validate, bool):
            raise TypeError("validate is expected to be a boolean")
        if scan == self.getScan():
            self.setCorScores(
                scan.saaxis_params.scores,
                score_method=scan.saaxis_params.score_method,
                img_width=scan.saaxis_params.image_width,
            )
            if scan.saaxis_params.autofocus is not None:
                self.setCurrentCorValue(scan.saaxis_params.autofocus)
            pm = ProcessManager()
            details = pm.get_dataset_details(
                dataset_id=scan.get_identifier(), process=self._processing_stack
            )
            current_state = pm.get_dataset_state(
                dataset_id=scan.get_identifier(), process=self._processing_stack
            )
            if current_state not in (
                DatasetState.CANCELLED,
                DatasetState.FAILED,
                DatasetState.SKIPPED,
            ):
                ProcessManager().notify_dataset_state(
                    dataset=scan,
                    process=self._processing_stack,
                    details=details,
                    state=DatasetState.WAIT_USER_VALIDATION,
                )
                self.sigResultsToBeShow.emit()
        if validate:
            _logger.processSucceed(
                f"saaxis processing succeeded with {scan.axis_params.relative_cor_value} as cor value"
            )
            self.validateScan(scan)

    def wait_processing(self, wait_time):
        self._processing_stack._computationThread.wait(wait_time)

    def validateCurrentScan(self):
        return self.validateScan(self.getScan())

    def validateScan(self, scan):
        if scan is None:
            return
        assert isinstance(scan, TomwerScanBase)
        selected_cor_value = self.getCurrentCorValue() or scan.saaxis_params.autofocus
        # if validate is done manually then pick current cor value; else we are in 'auto mode' and get it from the autofocus.
        details = ProcessManager().get_dataset_details(
            dataset_id=scan.get_identifier(), process=self._processing_stack
        )
        if details is None:
            details = ""
        if selected_cor_value is None:
            infos = f"no selected cor value. {scan} skip SAAXIS"
            infos = "\n".join((infos, details))
            _logger.warning(infos)
            scan.axis_params.set_relative_value(None)
            pv_helpers.notify_skip(
                process=self._processing_stack, dataset=scan, details=infos
            )
        else:
            scan.axis_params.set_relative_value(selected_cor_value)
            save_reconstruction_parameters_to_cache(scan=scan)

            if scan.nabu_recons_params is not None:
                if "reconstruction" not in scan.nabu_recons_params:
                    scan.nabu_recons_params["reconstruction"] = {}
                scan.nabu_recons_params["reconstruction"][
                    "rotation_axis_position"
                ] = scan.axis_params.absolute_cor_value
            r_cor = scan.axis_params.relative_cor_value
            a_cor = scan.axis_params.absolute_cor_value
            infos = f"cor selected for {scan}: relative: {r_cor}, absolute: {a_cor}"
            infos = "\n".join((infos, details))
            pv_helpers.notify_succeed(
                process=self._processing_stack, dataset=scan, details=infos
            )

        self.Outputs.data.send(scan)

    def getConfiguration(self) -> dict:
        config = super().getConfiguration()
        config["cluster_config"] = self._clusterConfig
        return config

    def setConfiguration(self, config: dict):
        # ignore slurm cluster. Defined by the upper widget
        config.pop("cluster_config", None)
        return super().setConfiguration(config)


class SAAxisOW(SuperviseOW, WidgetLongProcessing):
    """
    Widget for semi-automatic center of rotation calculation

    behavior within a workflow:

    * when a scan arrived:

        * if he already has a center of rotation defined (if the axis widget
          has been defined for example) it will be used as
          'estimated center of rotation'
        * if no cor has been computed yet and if the .nx entry contains
          information regarding an "x_rotation_axis_pixel_position" this value will
          be set

    * if autofocus option is lock:
        * launch the series of reconstruction (with research width defined)
          and the estimated center of rotation if defined. Once the
          reconstruction is ended and if the autofocus button is still lock
          it will select the cor with the highest
          value and mode to workflow downstream.

    * hint: you can define a "multi-step" half-automatic center of rotation
      research by creating several "saaxis" widget and reducing the research
      width.

    Details about :ref:`saaxis score calculation`
    """

    name = "multi-cor (sa-axis)"
    id = "orange.widgets.tomwer.sa_axis"
    description = "Reconstruct a slice with different center of rotation (cor) values"
    icon = "icons/saaxis.png"
    priority = 21
    keywords = [
        "multi",
        "multi-cor",
        "tomography",
        "semi automatic",
        "half automatic",
        "axis",
        "tomwer",
        "reconstruction",
        "rotation",
        "position",
        "center of rotation",
        "saaxis",
    ]

    ewokstaskclass = tomwer.core.process.reconstruction.saaxis.saaxis.SAAxisTask

    want_main_area = True
    resizing_enabled = True

    settingsHandler = CallbackSettingsHandler()

    sigScanReady = qt.Signal(TomwerScanBase)
    """Signal emitted when a scan is ready"""

    _ewoks_default_inputs = Setting({"data": None, "sa_axis_params": None})

    class Inputs:
        data = Input(
            name="data",
            type=TomwerScanBase,
            doc="one scan to be process",
            default=True,
            multiple=False,
        )
        data_recompute = Input(
            name="change recons params",
            type=_TomwerBaseDock,
        )
        cluster_in = Input(
            name="cluster_config",
            type=SlurmClusterConfiguration,
            doc="slurm cluster to be used",
            multiple=False,
        )

    class Outputs:
        data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")

    def __init__(self, parent=None):
        """

        :param parent: QWidget parent or None
        """
        SuperviseOW.__init__(self, parent)
        WidgetLongProcessing.__init__(self)
        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._widget = SAAxisWindow(
            Outputs=self.Outputs, parent=self, process_id=self.process_id
        )
        self._layout.addWidget(self._widget)

        sa_axis_params = self._ewoks_default_inputs.get("sa_axis_params", None)
        self.setConfiguration(sa_axis_params or {})

        # connect signal / slot
        self._widget.sigConfigurationChanged.connect(self._updateSettings)
        self._widget._processing_stack.sigComputationStarted.connect(
            self._startProcessing
        )
        self._widget._processing_stack.sigComputationEnded.connect(self._endProcessing)
        self._widget.sigValidated.connect(self.accept)
        self._widget.sigResultsToBeShow.connect(self._raiseResults)

        # expose API
        self.wait_processing = self._widget.wait_processing

    def getWaitingOverlay(self):
        return self._widget._tabWidget._resultsViewer._plot.getWaiterOverlay()

    def _startProcessing(self, *args, **kwargs):
        self.getWaitingOverlay().show()
        super()._startProcessing()

    def _endProcessing(self, *args, **kwargs):
        self.getWaitingOverlay().hide()
        super()._endProcessing()

    def __new__(cls, *args, **kwargs):
        # ensure backward compatibility with 'static_input'
        static_input = kwargs.get("stored_settings", {}).get("static_input", None)
        if static_input not in (None, {}):
            _logger.warning(
                "static_input has been deprecated. Will be replaced by _ewoks_default_inputs in the workflow file. Please save the workflow to apply modifications"
            )
            kwargs["stored_settings"]["_ewoks_default_inputs"] = static_input
        return super().__new__(cls, *args, **kwargs)

    def setConfiguration(self, configuration):
        if "workflow" in configuration:
            autofocus_lock = configuration["workflow"].get("autofocus_lock", None)
            if autofocus_lock is not None:
                self._widget.lockAutofocus(autofocus_lock)
            del configuration["workflow"]
        self._widget.setConfiguration(configuration)

    def getCurrentCorValue(self):
        return self._widget.getCurrentCorValue()

    def setEstimatedCorPosition(self, value):
        self._widget.setEstimatedCorPosition(value=value)

    def load_sinogram(self):
        self._widget.loadSinogram()

    def compute(self):
        self._widget.compute()

    def lockAutofocus(self, lock):
        self._widget.lockAutofocus(lock=lock)

    def isAutoFocusLock(self):
        return self._widget.isAutoFocusLock()

    @Inputs.data
    def process(self, scan):
        if scan is None:
            return
        if scan.axis_params is None:
            scan.axis_params = QAxisRP()
        if scan.saaxis_params is None:
            scan.saaxis_params = QSAAxisParams()
        self._skipCurrentScan(new_scan=scan)

        if settings.isOnLbsram(scan) and is_low_on_memory(settings.get_lbsram_path()):
            self.notify_skip(
                scan=scan,
                details=f"saaxis has been skiped for {scan} because of low space in lbsram",
            )
            self.Outputs.data.send(scan)
        else:
            self._widget.setScan(scan=scan)
            self.notify_pending(scan)
            self.activateWindow()
            if self.isAutoFocusLock():
                self.compute()
            else:
                self.raise_()
                self.show()

    @Inputs.cluster_in
    def setCluster(self, cluster):
        self._widget.setClusterConfig(cluster_config=cluster)

    def _skipCurrentScan(self, new_scan):
        scan = self._widget.getScan()
        # if the same scan has been run several scan
        if scan is None or str(scan) == str(new_scan):
            return
        current_scan_state = ProcessManager().get_dataset_state(
            dataset_id=scan.get_identifier(), process=self
        )
        if current_scan_state in (
            DatasetState.PENDING,
            DatasetState.WAIT_USER_VALIDATION,
            DatasetState.CANCELLED,
        ):
            details = "Was pending and has been replaced by another scan."
            self.notify_skip(scan=scan, details=details)
            self.Outputs.data.send(scan)

    @Inputs.data_recompute
    def reprocess(self, dataset):
        self.lockAutofocus(False)
        self.process(dataset)

    def cancel(self, scan):
        if scan is None:
            return
        if scan in self._widget._processing_stack:
            self._widget._processing_stack.remove(scan)
        if (
            self._widget._processing_stack._data_currently_computed.get_identifier().to_str()
            == scan.get_identifier().to_str()
        ):
            # stop current processing
            self._widget._processing_stack.cancel()
            # if possible process next
            if self._widget._processing_stack.can_process_next():
                self._widget._processing_stack._process_next()

    def validateCurrentScan(self):
        self._widget.validateCurrentScan()

    def _updateSettings(self):
        config = self._widget.getConfiguration()
        config.pop("cluster_config", None)

        self._ewoks_default_inputs = {
            "data": None,
            "sa_axis_params": self._widget.getConfiguration(),
        }
        self._ewoks_default_inputs["sa_axis_params"]["workflow"] = {
            "autofocus_lock": self._widget.isAutoFocusLock(),
        }

    def _raiseResults(self):
        if not self.isAutoFocusLock():
            self.raise_()
            self.show()
            self._widget.showResults()

    def getConfiguration(self):
        return self._widget.getConfiguration()



# coding: utf-8
from __future__ import annotations

import functools
import logging

from processview.core.manager import DatasetState, ProcessManager
from processview.core.superviseprocess import SuperviseProcess
from silx.gui import qt

from tomwer.core.process.reconstruction.axis import AxisTask
from tomwer.core.process.reconstruction.axis.axis import NoAxisUrl
from tomwer.core.process.reconstruction.axis.mode import AxisMode
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.settings import get_lbsram_path, isOnLbsram
from tomwer.core.utils.lbsram import is_low_on_memory
from tomwer.synctools.axis import QAxisRP

from ..processingstack import FIFO, ProcessingThread

_logger = logging.getLogger(__name__)


class AxisProcessStack(FIFO, qt.QObject):
    """Implementation of the `.AxisTask` but having a stack for treating
    scans and making computation in threads"""

    def __init__(self, axis_params, process_id=None):
        qt.QObject.__init__(self)
        FIFO.__init__(self, process_id=process_id)
        assert axis_params is not None
        self._axis_params = axis_params

    def _process(self, data, configuration, callback=None):
        ProcessManager().notify_dataset_state(
            dataset=data,
            process=self,
            state=DatasetState.ON_GOING,
        )
        assert isinstance(data, TomwerScanBase)
        if data.axis_params is None:
            data.axis_params = QAxisRP()
        self._data_currently_computed = data
        self._axis_params.frame_width = data.dim_1
        mode = self._axis_params.mode
        if isOnLbsram(data) and is_low_on_memory(get_lbsram_path()) is True:
            # if computer is running into low memory on lbsram skip it
            mess = "low memory, skip axis calculation", data.path
            ProcessManager().notify_dataset_state(
                dataset=data, process=self._process_id, state=DatasetState.SKIPPED
            )
            _logger.processSkipped(mess)
            data.axis_params.set_relative_value(None)
            if callback is not None:
                callback()
            self.scan_ready(scan=data)
        elif mode is (AxisMode.manual,):
            # if cor is not set then set it to 0 (can be the case if no)
            # interaction has been dne
            cor = self._axis_params.relative_cor_value
            if cor is None:
                cor = 0
            data._axis_params.set_relative_value(cor)
            cor = data._axis_params.relative_cor_value
            # If mode is read or manual the position_value is not computed and
            # we will keep the actual one (should have been defined previously)
            self._end_computation(data=data, future_tomo_obj=None, callback=callback)

        else:
            _logger.processStarted(
                f"Start cor calculation on {data} ({self._axis_params.get_simple_str()})"
            )
            data.axis_params.set_relative_value("...")
            self._axis_params.set_relative_value("...")
            assert self._axis_params.relative_cor_value == "..."
            self._data_currently_computed = data
            self._computationThread.init(data=data, axis_params=configuration)
            fct_callback = functools.partial(self._end_threaded_computation, callback)
            self._computationThread.finished.connect(fct_callback)
            self._computationThread.start()

    def _end_computation(self, data, future_tomo_obj, callback):
        """
        callback when the computation thread is finished

        :param scan: pass if no call to '_computationThread is made'
        """
        assert isinstance(data, TomwerScanBase)
        assert self._axis_params is not None
        # copy result computed on scan on the AxisTask reconsparams
        self._axis_params.set_relative_value(
            data.axis_params.relative_cor_value
        )  # noqa
        self._axis_params.frame_width = data.dim_1
        FIFO._end_computation(
            self, data=data, future_tomo_obj=future_tomo_obj, callback=callback
        )

    def _end_threaded_computation(self, callback=None):
        assert self._data_currently_computed is not None
        self._axis_params.set_relative_value(self._computationThread.center_of_rotation)
        self._computationThread.finished.disconnect()
        if callback:
            callback()
        FIFO._end_threaded_computation(self)

    def _create_processing_thread(self, process_id=None) -> qt.QThread:
        return _ProcessingThread(process_id=process_id)


class _ProcessingThread(ProcessingThread, SuperviseProcess):
    """
    Thread use to execute the processing of the axis position
    """

    def __init__(self, process_id=None):
        SuperviseProcess.__init__(self, process_id=process_id)
        try:
            ProcessingThread.__init__(self, process_id=process_id)
        except TypeError:
            ProcessingThread.__init__(self)
        self.center_of_rotation = None
        self._scan = None
        self._axis_params = None
        """function pointer to know which function to call for the axis
        calculation"""
        self.__patch = {}
        """Used to patch some calculation method (for test purpose)"""

    def init(self, data, axis_params):
        self._scan = data
        self._axis_params = axis_params

    def run(self):
        self.sigComputationStarted.emit()
        axis = AxisTask(
            inputs={
                "data": self._scan,
                "axis_params": self._axis_params,
                "serialize_output_data": False,
            },
            process_id=self.process_id,
        )
        try:
            axis.run()
        except NoAxisUrl as e:
            self.center_of_rotation = None
            _logger.error(f"CoR calculation failed. Issue with input ({e})")
        except Exception as e:
            _logger.error(f"CoR calculation failed ({e})", stack_info=True)
            self.center_of_rotation = None
        else:
            self.center_of_rotation = self._scan.axis_params.relative_cor_value



"""widgets to perform a z-sttiching in pre processing (from projections) or post processing (from reconstructed volumes)"""

from __future__ import annotations


import logging

from silx.gui import qt
from silx.gui.utils import blockSignals
from tomoscan.scanbase import TomoScanBase
from tomoscan.series import Series

from tomwer.io.utils.tomoobj import get_tomo_objs_instances
from tomwer.core.tomwer_object import TomwerObject
from tomwer.gui.stitching.metadataholder import QStitchingMetadata
from tomwer.gui.utils.illustrations import _IllustrationWidget
from .singleaxis import _SingleAxisMixIn

_logger = logging.getLogger(__name__)


class AxisOrderedTomoObjWidget(qt.QWidget, _SingleAxisMixIn):
    """
    main widget to define configuration of a z-stitching
    """

    sigConfigChanged = qt.Signal()
    """emit when the configuration changed"""
    sigAddTomoObjRequest = qt.Signal(TomwerObject)
    """request when a file is dropped or added through the 'add button'."""
    sigRemoveObjRequest = qt.Signal(TomwerObject)
    """request when an object should be removed by calling the 'remove button'."""

    def __init__(self, axis: int, parent=None):
        super().__init__(parent=parent)
        if axis not in (0, 1, 2):
            raise ValueError(f"axis should be in (0, 1, 2). Got {axis}")
        self._axis = axis  # implement _SingleAxisMixIn interface
        self.setLayout(qt.QGridLayout())

        self._addTomoObjCallbacks = tuple()
        self._removetomoObjCallbacks = tuple()

        # left panel with arrow
        img_flow = "flow_down"
        left_panel_width = 50
        axis_label = self.axis_alias(axis)
        self._axisLabel = qt.QLabel(axis_label, parent=self)
        self._axisLabel.setToolTip(f"{axis_label} is also know as axis {axis}")
        font = self._axisLabel.font()
        font.setPixelSize(40)
        self._axisLabel.setFont(font)
        self._axisLabel.setAlignment(qt.Qt.AlignCenter)
        self._axisLabel.setFixedWidth(left_panel_width)
        self.layout().addWidget(self._axisLabel, 1, 0, 1, 1)

        self._flowIllustration = _IllustrationWidget(parent=self, img=img_flow)
        self._flowIllustration.setFixedWidth(left_panel_width)
        self._flowIllustration.setSizePolicy(
            qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding
        )
        self.layout().addWidget(self._flowIllustration, 2, 0, 1, 1)
        self._flowIllustration.setToolTip(
            f"we are first ordering tomo object along axis {axis} (aka {self.axis_alias(axis)}) position to apply stitching"
        )

        # central object list
        self._tomoObjsTableView = qt.QTableView(parent=self)
        self._tomoObjsTableView.setModel(
            AxisOrderedTomoObjsModel(parent=self._tomoObjsTableView, axis=axis)
        )
        self._tomoObjsTableView.setSelectionBehavior(qt.QAbstractItemView.SelectRows)

        self._tomoObjsTableView.horizontalHeader().setStretchLastSection(True)
        self._tomoObjsTableView.setSortingEnabled(False)
        self._tomoObjsTableView.setColumnWidth(0, 20)
        self._tomoObjsTableView.setDragEnabled(True)
        self._tomoObjsTableView.setAcceptDrops(True)

        self.layout().addWidget(self._tomoObjsTableView, 2, 1, 1, 1)

    def addTomoObj(self, tomo_obj: TomoScanBase, trigger_callbacks=False):
        if tomo_obj is None:
            return
        else:
            self._tomoObjsTableView.model().addTomoObj(tomo_obj)
            # register tomo obj metadata modification to make sure we tell the z ordered list up to data
            tomo_obj.stitching_metadata.sigChanged.connect(self._orderMightHaveChanged)
            if trigger_callbacks:
                for callback in self._addTomoObjCallbacks:
                    callback(tomo_obj)

    def removeTomoObj(self, tomo_obj, trigger_callbacks=False):
        self._tomoObjsTableView.model().removeTomoObj(tomo_obj)
        tomo_obj.stitching_metadata.sigChanged.disconnect(self._orderMightHaveChanged)
        if trigger_callbacks:
            for callback in self._removetomoObjCallbacks:
                callback(tomo_obj)

    def _orderMightHaveChanged(self):
        # when position over {axis} is updated from the GUI, make sure the {axis} ordered list of tomo object is still
        # ordered and selection is still the accurate one.
        # this case is not handled on this widget
        self._tomoObjsTableView.model().reorder_objs()

    def getTomoObjsAxisOrdered(self) -> tuple:
        return tuple(self._tomoObjsTableView.model()._axis_decreasing_ordered_objs)

    def clearTomoObjs(self):
        self._tomoObjsTableView.model().clearTomoObjs()


class AxisOrderedTomoObjsModel(qt.QAbstractTableModel):
    def __init__(self, axis: int, parent=None) -> None:
        super().__init__(parent)
        assert axis in (0, 1, 2)
        self._axis = axis
        # either we plot the tomo_obj in the list within the full identifier or with the reduce str (in this case two scans can have identical name. Identifier is unique)
        self._objs = Series(name="unordered tomo obj", use_identifiers=False)
        self._axis_decreasing_ordered_objs = []
        self._headers = ["index", "tomo obj"]
        self._tomoObjToIndex = {}
        # for each tomo object store a 'unique' id
        self._nextIndex = -1

    def supportedDragActions(self):
        return qt.Qt.CopyAction

    def supportedDropActions(self):
        return qt.Qt.CopyAction

    def getTomoObjs(self) -> tuple:
        return tuple(self._objs)

    def pop_index(self):
        self._nextIndex += 1
        return self._nextIndex

    def getTomoObjCurrentPos(self, tomo_obj):
        try:
            index = self._axis_decreasing_ordered_objs.index(tomo_obj)
        except ValueError:  # if not in the list
            return -1
        else:
            return index

    def clearTomoObjs(self):
        self._objs = Series(name="unordered tomo obj", use_identifiers=False)
        self._axis_decreasing_ordered_objs = []
        self.layoutChanged.emit()

    def reorder_objs(self):
        def get_min_axis_or_0(tomo_obj):
            # expects to find a position. If cannot set the value to infinity to be sure all tomo objects without
            # metadata (raw or from the user) are at the specific position
            return (
                tomo_obj.stitching_metadata.get_abs_position_px(axis=self._axis) or 0.0
            )

        self._axis_decreasing_ordered_objs = Series(
            "ordered tomo obj",
            sorted(self._objs[:], key=get_min_axis_or_0, reverse=True),
            use_identifiers=False,
        )

    def addTomoObj(self, obj: TomwerObject):
        if not isinstance(obj, TomwerObject):
            raise TypeError(
                f"{obj} is expected to be an instance of {TomwerObject}. {type(obj)} provided instead"
            )
        elif obj.stitching_metadata is None:
            # make sure it can contain some stitching metadata
            obj.stitching_metadata = QStitchingMetadata(tomo_obj=obj)
        elif not isinstance(obj.stitching_metadata, QStitchingMetadata):
            # if need to upgrade it to a QObject
            obj.stitching_metadata = QStitchingMetadata.from_dict(
                obj.stitching_metadata.to_dict(), tomo_obj=obj
            )
        self._objs.append(obj)
        identifier = obj.get_identifier().to_str()
        if identifier not in self._tomoObjToIndex:
            self._tomoObjToIndex[identifier] = self.pop_index()
        self.reorder_objs()
        self.layoutChanged.emit()

    def removeTomoObj(self, obj: TomwerObject):
        if not isinstance(obj, TomwerObject):
            raise TypeError(
                f"{obj} is expected to be an instance of {TomwerObject}. {type(obj)} provided instead"
            )
        self._objs.remove(obj)
        # note: we avoid removing the identifier as this is pretty small and convenient to keep
        # in case of wromg manipulation when removing an item
        self.reorder_objs()
        self.layoutChanged.emit()

    def rowCount(self, parent=None):
        return len(self._objs)

    def columnCount(self, parent=None):
        return len(self._headers)

    def headerData(self, col, orientation, role):
        if orientation == qt.Qt.Horizontal and role == qt.Qt.DisplayRole:
            if col < len(self._headers):
                return self._headers[col]
            else:
                return None

    def getTomoObj(self, qmodelindex: qt.QModelIndex):
        index = qmodelindex.row()
        if index < len(self._axis_decreasing_ordered_objs):
            return self._axis_decreasing_ordered_objs[index]

    def data(self, index, role):
        if not index.isValid():
            return None

        if index.column() == 0:
            if role == qt.Qt.TextAlignmentRole:
                return qt.Qt.AlignLeft | qt.Qt.AlignVCenter
            elif role == qt.Qt.DisplayRole:
                obj = self._axis_decreasing_ordered_objs[index.row()]
                identifier = obj.get_identifier().to_str()
                return self._tomoObjToIndex.get(identifier, "???")
        elif index.column() == 1:
            if role == qt.Qt.TextAlignmentRole:
                return qt.Qt.AlignHCenter | qt.Qt.AlignVCenter
            elif role == qt.Qt.ToolTipRole:
                obj = self._axis_decreasing_ordered_objs[index.row()]
                return obj.get_identifier().to_str()
            elif role == qt.Qt.DisplayRole:
                obj = self._axis_decreasing_ordered_objs[index.row()]
                return obj.get_identifier().short_description()
            else:
                return None
        else:
            return None


class AxisOrderedTomoObjWidgetSingleSel(AxisOrderedTomoObjWidget):
    """
    AxisOrderedTomoObjWidget with a list that can select a single line.

    The idea is that it can be used to select the object to be edited on another widget
    """

    def __init__(self, axis: int, parent=None):
        super().__init__(axis=axis, parent=parent)
        self._tomoObjsTableView.setSelectionMode(qt.QAbstractItemView.SingleSelection)

    def removeTomoObj(self, tomo_obj):
        selected = self.getSelectedTomoObj()
        super().removeTomoObj(tomo_obj=tomo_obj)
        if selected is tomo_obj:
            self.setSelectedTomoObj(None)
        else:
            self.setSelectedTomoObj(selected)

    def getSelectedTomoObj(self):
        selection = self._tomoObjsTableView.selectedIndexes()
        if len(selection) > 0:
            return self._tomoObjsTableView.model().getTomoObj(selection[0])
        return None

    def setSelectedTomoObj(self, tomo_obj: TomwerObject | None):
        self._tomoObjsTableView.selectionModel().clearSelection()
        if tomo_obj is not None:
            model = self._tomoObjsTableView.model()
            tomo_obj_ordered_row = model.getTomoObjCurrentPos(tomo_obj=tomo_obj)

            if tomo_obj_ordered_row >= 0:
                for i_column in range(model.columnCount()):
                    selection = model.createIndex(
                        tomo_obj_ordered_row,
                        i_column,
                    )
                    self._tomoObjsTableView.selectionModel().select(
                        selection,
                        qt.QItemSelectionModel.Select,
                    )

    def _orderMightHavechanged(self):
        # when z position is updated from the GUI, make sure the z ordered list of tomo object is still
        # ordered and selection is still the accurate one.
        with blockSignals(self._tomoObjsTableView.selectionModel()):
            selectedTomoObj = self.getSelectedTomoObj()
            self._tomoObjsTableView.model().reorder_objs()
            self._tomoObjsTableView.model().layoutChanged.emit()
            self.setSelectedTomoObj(selectedTomoObj)


class EditableOrderedTomoObjWidget(AxisOrderedTomoObjWidget):
    """
    same as the ZOrderedTomoObjWidget but you can add and remove tomo obj to the list
    """

    def __init__(self, axis: int, parent=None):
        super().__init__(axis=axis, parent=parent)
        if axis not in (0, 1, 2):
            raise ValueError(f"axis must be in (0, 1, 2). Got {axis}")
        self._axis = axis
        self.setAcceptDrops(True)

        # allow multiple selection as we might want to remove several object at the same time
        self._tomoObjsTableView.setSelectionMode(qt.QAbstractItemView.MultiSelection)

        self._buttons = qt.QWidget(self)
        self._buttons.setLayout(qt.QHBoxLayout())

        # allow drag and drop of elmts

        # create buttons
        ## spacer
        self._spacer = qt.QWidget(self._buttons)
        self._spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self._buttons.layout().addWidget(self._spacer)
        ## add button
        self._addButton = qt.QPushButton("Add")
        self._addButton.clicked.connect(self._callbackAddTomoObj)
        self._buttons.layout().addWidget(self._addButton)
        ## rm button
        self._rmButton = qt.QPushButton("Remove")
        self._rmButton.clicked.connect(self._callbackRemoveSelectedTomoObj)
        self._buttons.layout().addWidget(self._rmButton)

        self.layout().addWidget(self._buttons, 999, 0, 1, 2)

    @property
    def axis(self) -> int:
        return self._axis

    def removeTomoObj(self, tomo_obj, trigger_callbacks=False):
        selected = set(self.getSelectedTomoObjs())
        super().removeTomoObj(tomo_obj=tomo_obj, trigger_callbacks=trigger_callbacks)
        if tomo_obj in selected:
            selected.remove(tomo_obj)
            self.setSelectedTomoObjs(selected)

    def setSelectedTomoObjs(self, tomo_objs):
        self._tomoObjsTableView.clearSelection()

        model = self._tomoObjsTableView.model()
        for tomo_obj in tomo_objs:
            tomo_obj_ordered_row = model.getTomoObjCurrentPos(tomo_obj=tomo_obj)

            if tomo_obj_ordered_row >= 0:
                for i_column in range(model.columnCount()):
                    selection = model.createIndex(
                        tomo_obj_ordered_row,
                        i_column,
                    )
                    self._tomoObjsTableView.selectionModel().select(
                        selection,
                        qt.QItemSelectionModel.Select,
                    )

    def getSelectedTomoObjs(self):
        selection = self._tomoObjsTableView.selectedIndexes()
        return set(
            [self._tomoObjsTableView.model().getTomoObj(item) for item in selection]
        )

    def _callbackAddTomoObj(self) -> tuple[TomwerObject]:
        dialog = qt.QFileDialog()
        dialog.setFileMode(qt.QFileDialog.ExistingFiles)
        dialog.setNameFilters(
            [
                "Any file (*)",
            ]
        )

        if not dialog.exec():
            dialog.close()
            return ()
        elif len(dialog.selectedFiles()) == 0:
            return ()
        else:
            tomo_objs = []
            for file in dialog.selectedFiles():
                try:
                    new_objs = get_tomo_objs_instances(tomo_objs=(file,))[0]
                except Exception:
                    pass
                else:
                    tomo_objs.extend(new_objs)
            for tomo_obj in tomo_objs:
                self.addTomoObj(tomo_obj=tomo_obj, trigger_callbacks=True)
            return tuple(tomo_objs)

    def _callbackRemoveSelectedTomoObj(self):
        obj_to_remove = self.getSelectedTomoObjs()
        for obj in obj_to_remove:
            self.removeTomoObj(tomo_obj=obj, trigger_callbacks=True)

    def dropEvent(self, a0) -> None:
        if a0.mimeData().hasFormat("text/uri-list"):
            paths = [url.path() for url in a0.mimeData().urls()]
            tomo_objs = get_tomo_objs_instances(paths)[0]

            for tomo_obj in tomo_objs:
                assert isinstance(
                    tomo_obj, TomwerObject
                ), f"expected type is tomo_obj. Get {type(tomo_obj)}"
                self.addTomoObj(tomo_obj=tomo_obj, trigger_callbacks=True)
        else:
            return super().dropEvent(a0)

    def supportedDropActions(self):
        """Inherited method to redefine supported drop actions."""
        return qt.Qt.CopyAction | qt.Qt.MoveAction

    def dragEnterEvent(self, event):
        if event.mimeData().hasFormat("text/uri-list"):
            event.accept()
            event.setDropAction(qt.Qt.CopyAction)
        else:
            qt.QListWidget.dragEnterEvent(self, event)

    def dragMoveEvent(self, event):
        if event.mimeData().hasFormat("text/uri-list"):
            event.setDropAction(qt.Qt.CopyAction)
            event.accept()
        else:
            qt.QListWidget.dragMoveEvent(self, event)

    def setAddTomoObjCallbacks(self, callbacks: tuple):
        """
        To synchronize the different widget we might need to notify other widget that a tomo_obj has been added
        from the dedicated interface.
        Callbacks must take a single TomwerObject as input
        """
        self._addTomoObjCallbacks = callbacks

    def setRemoveTomoObjCallbacks(self, callbacks: tuple):
        """
        To synchronize the different widget we might need to notify other widget that a tomo_obj has been removed
        from the dedicated interface.
        Callbacks must take a single TomwerObject as input
        """
        self._removetomoObjCallbacks = callbacks

    def _orderMightHavechanged(self):
        # when z position is updated from the GUI, make sure the z ordered list of tomo object is still
        # ordered and selection is still the accurate one.
        with blockSignals(self._tomoObjsTableView.selectionModel()):
            selectedTomoObjs = self.getSelectedTomoObjs()
            self._tomoObjsTableView.model().reorder_objs()
            self._tomoObjsTableView.model().layoutChanged.emit()
            self.setSelectedTomoObjs(selectedTomoObjs)



# coding: utf-8
from __future__ import annotations

import unittest

from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from tomwer.gui.visualization.dataviewer import DataViewer, ImageStack


class TestImageStack(TestCaseQt):
    """Test that the ImageStackWidget can be load and remove without any issue"""

    def setUp(self):
        TestCaseQt.setUp(self)
        self._widget = ImageStack(None)

    def tearDown(self):
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        self.qapp.processEvents()
        unittest.TestCase.tearDown(self)

    def test(self):
        pass


class TestDataViewer(TestCaseQt):
    """Test that the data viewer can be load and remove without any issue"""

    def setUp(self):
        TestCaseQt.setUp(self)
        self._widget = DataViewer(None)

    def tearDown(self):
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        self.qapp.processEvents()
        unittest.TestCase.tearDown(self)

    def test(self):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Application to browse a given scan"""

import numpy
import argparse
import logging
import signal
import sys

import silx
from silx.gui import qt

from tomwer.tasks.visualization.volume_viewer import VolumeViewerTask
from tomwer.core.volume.volumefactory import VolumeFactory
from tomoscan.esrf.volume.utils import guess_volumes
from tomwer.core.utils.resource import increase_max_number_file
from tomwer.gui import icons
from tomwer.gui.utils.splashscreen import getMainSplashScreen
from tomwer.gui.visualization.volume_viewer.VolumeViewerWindow import (
    VolumeViewerWindow,
)

logging.basicConfig(level=logging.WARNING)
_logger = logging.getLogger(__name__)


class VolumeSummaryTaskThread(qt.QThread):
    def __init__(self, parent, inputs):
        super().__init__(parent)
        self._inputs = inputs

    def run(self):
        self._task = VolumeViewerTask(inputs=self._inputs)
        self._task.execute()

    def result(self):
        return {
            "slices": self._task.outputs.slices,
            "metadata": self._task.outputs.volume_metadata,
            "volume_shape": self._task.outputs.volume_shape,
        }

    def get_loaded_volume_data(self) -> numpy.array:
        return self._task.outputs.loaded_volume


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "volume",
        help="Volume identifier or location of the volume to display",
    )
    parser.add_argument(
        "--load-volume",
        help="Load the full volume in memory. Allows fast browsing along the three axis.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--use-opengl-plot",
        help="Use OpenGL for plots (instead of matplotlib)",
        action="store_true",
        default=False,
    )

    options = parser.parse_args(argv[1:])

    try:
        volumes = VolumeFactory.create_tomo_object_from_identifier(options.volume)
    except Exception:
        volumes = []

    if len(volumes) == 0:
        volumes = guess_volumes(
            options.volume,
        )

    if len(volumes) == 0:
        raise ValueError(f"No volume found in {options.volume}")
    if len(volumes) > 1:
        _logger.warning(
            f"More than one volume found in {options.volume}. Pick the first one."
        )

    volume = VolumeFactory.create_tomo_object_from_identifier(
        volumes[0].get_identifier().to_str()
    )

    if options.use_opengl_plot:
        silx.config.DEFAULT_PLOT_BACKEND = "gl"
    else:
        silx.config.DEFAULT_PLOT_BACKEND = "matplotlib"

    increase_max_number_file()

    global app  # QApplication must be global to avoid seg fault on quit

    app = qt.QApplication.instance() or qt.QApplication(["tomwer"])
    splash = getMainSplashScreen()
    qt.QApplication.setOverrideCursor(qt.Qt.WaitCursor)
    qt.QApplication.processEvents()

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler

    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catch
    timer.timeout.connect(lambda: None)

    window = VolumeViewerWindow(parent=None)
    # there is options from the CLI to load or not the volume at start. This option has no sense
    # in the CLI tool.
    window.setLoadingVolumeOptionVisible(False)
    window.setWindowTitle("tomwer: volume-viewer")
    window.setWindowIcon(icons.getQIcon("tomwer"))

    if options.load_volume:
        loading_message = "Loading volume."
    else:
        loading_message = "Loading slices."
    window.initVolumePreview(volume, message=loading_message)
    splash.finish(window)
    window.show()

    taskThread = VolumeSummaryTaskThread(
        parent=window,
        inputs={
            "volume": volume,
            "load_volume": options.load_volume,
        },
    )
    taskThread.finished.connect(
        lambda: window.setSlicesAndMetadata(**taskThread.result())
    )
    if options.load_volume:
        taskThread.finished.connect(
            lambda: window.setVolume(taskThread.get_loaded_volume_data())
        )
    taskThread.finished.connect(lambda: window.setLoading(False))
    taskThread.start()

    qt.QApplication.restoreOverrideCursor()
    app.aboutToQuit.connect(window.close)
    exit(app.exec())


def getinputinfo():
    return "tomwer volume-viewer [file_path] [[file_entry]]"


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.closeAllWindows()  # needed because get a waiting thread behind
    qt.QApplication.quit()


if __name__ == "__main__":
    main(sys.argv)



# coding: utf-8
from __future__ import annotations

import logging
import os
import shutil
import tempfile
import unittest
from glob import glob

import pytest
from nxtomomill.converter import from_h5_to_nx
from nxtomomill.models.h52nx import H52nxModel
from silx.gui.utils.testutils import TestCaseQt

from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils.spec import rebaseParFile
from tomwer.core.utils.scanutils import MockEDF
from tomwer.synctools.datatransfert import ScanTransfer
from tomwer.tests.datasets import TomwerCIDatasets

try:
    from tomwer.synctools.rsyncmanager import RSyncManager
except ImportError:
    has_rsync = False
else:
    has_rsync = True

logging.disable(logging.INFO)


class TestEDFDataTransfert(TestCaseQt):
    """
    Test that the folder transfert process is valid
    """

    def setUp(self):
        TestCaseQt.setUp(self)
        self.sourcedir = tempfile.mkdtemp()
        self.n_file = 10
        MockEDF.fastMockAcquisition(self.sourcedir, n_radio=self.n_file)
        self.scan = ScanFactory.create_scan_object(self.sourcedir)
        assert os.path.isdir(self.sourcedir)
        assert os.path.exists(self.sourcedir)
        self.targettedir = tempfile.mkdtemp()
        assert os.path.isdir(self.targettedir)
        assert os.path.exists(self.targettedir)
        self.outputdir = os.path.join(
            self.targettedir, os.path.basename(self.sourcedir)
        )

    def tearDown(self):
        if os.path.isdir(self.sourcedir):
            shutil.rmtree(self.sourcedir)
        if os.path.isdir(self.targettedir):
            shutil.rmtree(self.targettedir)

    def testMoveFiles(self):
        """
        simple test that files are moved
        """
        folder_trans_process = ScanTransfer(
            inputs={
                "data": self.scan,
                "turn_off_print": True,
                "copying": False,
                "dest_dir": self.targettedir,
                "noRsync": True,
            }
        )
        folder_trans_process.run()
        self.assertTrue(os.path.isdir(self.outputdir))
        self.assertTrue(self.checkDataCopied())

    def testCopyFiles(self):
        """
        Simple test that file are copy and deleted
        """
        folder_trans_process = ScanTransfer(
            inputs={
                "data": self.scan,
                "turn_off_print": True,
                "copying": False,
                "dest_dir": self.targettedir,
                "noRsync": True,
                "move": False,
            }
        )
        folder_trans_process.run()
        self.assertTrue(self.checkDataCopied())

    def testMoveFilesForce(self):
        """
        Test the force option of folderTransfert
        """
        assert not os.path.isdir(self.outputdir)
        assert os.path.isdir(self.scan.path)
        folder_trans_process = ScanTransfer(
            inputs={
                "data": self.scan,
                "turn_off_print": True,
                "copying": False,
                "dest_dir": self.targettedir,
                "noRsync": True,
                "move": True,
                "overwrite": False,
            }
        )
        folder_trans_process.run()

        MockEDF.fastMockAcquisition(self.sourcedir, n_radio=self.n_file)
        folder_trans_process = ScanTransfer(
            inputs={
                "data": self.scan,
                "turn_off_print": True,
                "copying": False,
                "dest_dir": self.targettedir,
                "noRsync": True,
                "move": True,
                "overwrite": False,
            }
        )
        with self.assertRaises(shutil.Error):
            self.assertRaises(folder_trans_process.run())

        folder_trans_process = ScanTransfer(
            inputs={
                "data": self.scan,
                "turn_off_print": True,
                "copying": False,
                "dest_dir": self.targettedir,
                "noRsync": True,
                "move": True,
                "overwrite": True,
            }
        )
        folder_trans_process.run()
        self.assertTrue(self.checkDataCopied())

    def testCopyFilesForce(self):
        """
        Test the force option for the copy files process
        """
        assert not os.path.isdir(self.outputdir)
        os.makedirs(self.outputdir)
        assert os.path.isdir(self.outputdir)

        folder_trans_process = ScanTransfer(
            inputs={
                "data": self.scan,
                "turn_off_print": True,
                "copying": False,
                "dest_dir": self.targettedir,
                "noRsync": True,
                "move": False,
                "overwrite": False,
            }
        )
        folder_trans_process.run()
        self.assertTrue(self.checkDataCopied())

        MockEDF.fastMockAcquisition(self.sourcedir, n_radio=self.n_file)
        self.assertTrue(self.scan.path == self.sourcedir)

        folder_trans_process = ScanTransfer(
            inputs={
                "data": self.scan,
                "turn_off_print": True,
                "copying": False,
                "dest_dir": self.targettedir,
                "noRsync": True,
                "move": False,
                "overwrite": False,
            }
        )

        with self.assertRaises(FileExistsError):
            self.assertRaises(folder_trans_process.run())

        folder_trans_process = ScanTransfer(
            inputs={
                "data": self.scan,
                "turn_off_print": True,
                "copying": False,
                "dest_dir": self.targettedir,
                "noRsync": True,
                "move": False,
                "overwrite": True,
            }
        )
        folder_trans_process.run()
        self.assertTrue(self.checkDataCopied())

    def checkDataCopied(self):
        outputFiles = os.listdir(self.outputdir)
        inputFile = glob(self.sourcedir)
        # + 3 because .info and .xml are count
        return (
            (len(inputFile) == 0)
            and (len(outputFiles) == (self.n_file + 2))
            and (not os.path.isdir(self.sourcedir))
        )


@pytest.mark.skipif(RSyncManager().has_rsync() is False, reason="Rsync is missing")
class TestHDFDataTransfert(TestCaseQt):
    """Make sure we can transfer data from bliss acquisition"""

    def setUp(self):
        super().setUp()
        self.input_dir = tempfile.mkdtemp()
        self.output_dir = tempfile.mkdtemp()
        shutil.copytree(
            TomwerCIDatasets.get_dataset("bliss/sample"),
            os.path.join(self.input_dir, "sample"),
        )

        self._proposal_file = os.path.join(
            self.input_dir, "sample", "ihpayno_sample.h5"
        )
        assert os.path.exists(self._proposal_file)
        self._sample_file = os.path.join(
            self.input_dir, "sample", "sample_29042021", "sample_29042021.h5"
        )
        assert os.path.exists(self._sample_file)
        self._sample_file_entry = "/1.1"

        output_file_path = os.path.join(
            self.input_dir, "sample", "sample_29042021", "nx_file.nx"
        )
        # convert it to nx
        configuration = H52nxModel()
        configuration.input_file = self._sample_file
        configuration.output_file = output_file_path
        files_entries = (self._sample_file_entry,)
        configuration.entries = files_entries
        configuration.single_file = False
        configuration.overwrite = False
        configuration.request_input = False
        configuration.file_extension = ".nx"
        results = from_h5_to_nx(configuration)

        assert len(results) == 1
        output_file_path, entry = results[0]
        assert os.path.exists(output_file_path)
        assert os.path.isfile(output_file_path)
        self.scan = NXtomoScan(scan=output_file_path, entry=entry)

    def tearDown(self):
        for dir_ in (self.input_dir, self.output_dir):
            shutil.rmtree(dir_)

    def testDataTransfert(self):
        process = ScanTransfer(
            inputs={
                "data": self.scan,
                "dest_dir": os.path.join(self.output_dir, "sample"),
                "block": True,
            }
        )
        process.run()
        output_scan = os.path.join(
            self.output_dir, "sample", "sample_29042021", "scan0002"
        )
        input_scan = os.path.join(
            self.input_dir, "sample", "sample_29042021", "scan0002"
        )
        output_proposal_file = os.path.join(
            self.output_dir, "sample", "ihpayno_sample.h5"
        )
        output_sample_file = os.path.join(
            self.output_dir, "sample", "sample_29042021", "sample_29042021.h5"
        )
        self.assertTrue(os.path.exists(output_scan))
        self.assertEqual(len(os.listdir(output_scan)), len(os.listdir(input_scan)))
        self.assertTrue(os.path.exists(output_proposal_file))
        self.assertTrue(os.path.exists(output_sample_file))
        other_scan_dir = os.path.join(self.output_dir, "sample", "scan0004")
        self.assertFalse(os.path.exists(other_scan_dir))


class TestPreTransfert(unittest.TestCase):
    """Test the pretransfert actions"""

    def setUp(self):
        unittest.TestCase.setUp(self)
        folderDataset = TomwerCIDatasets.get_dataset("edf_datasets/scan_3_")
        self.tmpdir = tempfile.mkdtemp()
        self.outputfolder = tempfile.mkdtemp()
        scan_path = os.path.join(self.tmpdir, "scan_3_")
        shutil.copytree(src=folderDataset, dst=scan_path)
        self.scan = ScanFactory.create_scan_object(scan_path=scan_path)

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        for _dir in (self.tmpdir, self.outputfolder):
            if os.path.exists(_dir):
                shutil.rmtree(_dir)

    def testParFile(self):
        """Make sure the `scan_3_.par` is correctly updated when moved"""
        parDict = self.getParDict(os.path.join(self.scan.path, "scan_3_.par"))
        assert (
            parDict["file_prefix"]
            == "/data/visitor/mi1226/id19/nemoz/henri/scan_3_/scan_3_"
        )
        assert (
            parDict["ff_prefix"]
            == "/data/visitor/mi1226/id19/nemoz/henri/scan_3_/refHST"
        )
        rebaseParFile(
            os.path.join(self.scan.path, "scan_3_.par"),
            oldfolder="/data/visitor/mi1226/id19/nemoz/henri/scan_3_",
            newfolder=self.scan.path,
        )
        parDict = self.getParDict(os.path.join(self.scan.path, "scan_3_.par"))
        self.assertTrue(
            parDict["file_prefix"] == os.path.join(self.scan.path, "scan_3_")
        )
        self.assertTrue(parDict["ff_prefix"] == os.path.join(self.scan.path, "refHST"))
        folderTrans = ScanTransfer(
            inputs={
                "turn_off_print": True,
                "copying": False,
                "dest_dir": self.outputfolder,
                "data": self.scan,
                "move": False,
                "overwrite": True,
                "noRsync": True,
            }
        )

        folderTrans.run()
        parDict = self.getParDict(
            os.path.join(self.outputfolder, "scan_3_", "scan_3_.par")
        )
        self.assertTrue(
            parDict["file_prefix"]
            == os.path.join(self.outputfolder, "scan_3_", "scan_3_")
        )
        self.assertTrue(
            parDict["ff_prefix"] == os.path.join(self.outputfolder, "scan_3_", "refHST")
        )

    @staticmethod
    def getParDict(_file):
        assert os.path.isfile(_file)
        ddict = {}
        f = open(_file, "r")
        lines = f.readlines()
        for line in lines:
            if "=" not in line:
                continue
            line_str = line.rstrip().replace(" ", "")
            line_str = line_str.split("#")[0]
            key, value = line_str.split("=")
            ddict[key.lower()] = value
        return ddict



# coding: utf-8
from __future__ import annotations


import os
import shutil
import tempfile
import unittest

import h5py
import numpy
from silx.io.url import DataUrl
from silx.io.utils import h5py_read_dataset
from tomoscan.esrf.scan.mock import MockNXtomo

from tomwer.core.process.edit.darkflatpatch import DarkFlatPatchTask
from tomwer.core.scan.nxtomoscan import NXtomoScan


class BaseTestAddDarkAndFlats(unittest.TestCase):
    """
    Unit test on nxtomomill.utils.add_dark_flat_nx_file function
    """

    def setUp(self) -> None:
        self.tmpdir = tempfile.mkdtemp()
        simple_nx_path = os.path.join(self.tmpdir, "simple_case")
        self.dim = 55
        self.nproj = 20
        self._simple_nx = MockNXtomo(
            scan_path=simple_nx_path,
            n_proj=self.nproj,
            n_ini_proj=self.nproj,
            create_ini_dark=False,
            create_ini_flat=False,
            create_final_flat=False,
            dim=self.dim,
        ).scan
        with h5py.File(self._simple_nx.master_file, mode="r") as h5s:
            data_path = "/".join(
                (self._simple_nx.entry, "instrument", "detector", "data")
            )
            self._raw_data = h5py_read_dataset(h5s[data_path])
        nx_with_vds_path = os.path.join(self.tmpdir, "case_with_vds")
        self._nx_with_virtual_dataset = MockNXtomo(
            scan_path=nx_with_vds_path,
            n_proj=0,
            n_ini_proj=0,
            create_ini_dark=False,
            create_ini_flat=False,
            create_final_flat=False,
            dim=self.dim,
        ).scan
        self._create_vds(
            source_file=self._simple_nx.master_file,
            source_data_path=self._simple_nx.entry,
            target_file=self._nx_with_virtual_dataset.master_file,
            target_data_path=self._nx_with_virtual_dataset.entry,
            copy_other_data=True,
        )
        self._patch_nxtomo_flags(self._nx_with_virtual_dataset)

        nx_with_vds_path_and_links = os.path.join(
            self.tmpdir, "case_with_vds_and_links"
        )
        self._nx_with_virtual_dataset_with_link = MockNXtomo(
            scan_path=nx_with_vds_path_and_links,
            n_proj=0,
            n_ini_proj=0,
            create_ini_dark=False,
            create_ini_flat=False,
            create_final_flat=False,
            dim=self.dim,
        ).scan
        self._create_vds(
            source_file=self._simple_nx.master_file,
            source_data_path=self._simple_nx.entry,
            target_file=self._nx_with_virtual_dataset_with_link.master_file,
            target_data_path=self._nx_with_virtual_dataset_with_link.entry,
            copy_other_data=True,
        )
        self._patch_nxtomo_flags(self._nx_with_virtual_dataset_with_link)

        # create dark
        self.start_dark = (
            numpy.random.random((self.dim * self.dim))
            .reshape(1, self.dim, self.dim)
            .astype("f")
        )
        self.start_dark_file = os.path.join(self.tmpdir, "dark.hdf5")
        self.start_dark_entry = "data"
        self.start_dark_url = self._save_raw(
            data=self.start_dark,
            entry=self.start_dark_entry,
            file_path=self.start_dark_file,
        )

        self.end_dark = (
            numpy.random.random((self.dim * self.dim * 2))
            .reshape(2, self.dim, self.dim)
            .astype("f")
        )
        self.end_dark_file = os.path.join(self.tmpdir, "dark.hdf5")
        self.end_dark_entry = "data2"
        self.end_dark_url = self._save_raw(
            data=self.end_dark, entry=self.end_dark_entry, file_path=self.end_dark_file
        )

        # create flats
        self.start_flat = (
            numpy.random.random((self.dim * self.dim * 3))
            .reshape(3, self.dim, self.dim)
            .astype("f")
        )
        self.start_flat_file = os.path.join(self.tmpdir, "start_flat.hdf5")
        self.start_flat_entry = "/root/flat"
        self.start_flat_url = self._save_raw(
            data=self.start_flat,
            entry=self.start_flat_entry,
            file_path=self.start_flat_file,
        )

        self.end_flat = (
            numpy.random.random((self.dim * self.dim))
            .reshape(1, self.dim, self.dim)
            .astype("f")
        )
        # save the end flat in the simple case file to insure all cases are
        # consider
        self.end_flat_file = self._simple_nx.master_file
        self.end_flat_entry = "flat"
        self.end_flat_url = self._save_raw(
            data=self.end_flat, entry=self.end_flat_entry, file_path=self.end_flat_file
        )

    def _save_raw(self, data, entry, file_path) -> DataUrl:
        with h5py.File(file_path, mode="a") as h5s:
            h5s[entry] = data
        return DataUrl(file_path=file_path, data_path=entry, scheme="silx")

    def _create_vds(
        self,
        source_file: str,
        source_data_path: str,
        target_file: str,
        target_data_path: str,
        copy_other_data: bool,
    ):
        """Create virtual dataset and links from source to target

        :param source_file:
        :param source_data_path:
        :param target_file:
        :param target_data_path:
        :param copy_other_data: we want to create two cases: one copying
                                     datasets 'image_key'... and the other
                                     one linking them. Might have a difference
                                     of behavior when overwriting for example
        """
        assert source_file != target_file, "file should be different"
        # link data
        n_frames = 0
        # for now we only consider the original data
        with h5py.File(source_file, mode="r") as o_h5s:
            old_path = os.path.join(source_data_path, "instrument", "detector", "data")
            n_frames += o_h5s[old_path].shape[0]
            shape = o_h5s[old_path].shape
            data_type = o_h5s[old_path].dtype

            layout = h5py.VirtualLayout(shape=shape, dtype=data_type)
            assert os.path.exists(source_file)
            with h5py.File(source_file, mode="r") as ppp:
                assert source_data_path in ppp
            layout[:] = h5py.VirtualSource(path_or_dataset=o_h5s[old_path])

            det_path = os.path.join(target_data_path, "instrument", "detector")
            with h5py.File(target_file, mode="a") as h5s:
                detector_node = h5s.require_group(det_path)
                detector_node.create_virtual_dataset("data", layout, fillvalue=-5)

        for path in (
            os.path.join("instrument", "detector", "image_key"),
            os.path.join("instrument", "detector", "image_key_control"),
            os.path.join("instrument", "detector", "count_time"),
            os.path.join("sample", "rotation_angle"),
        ):
            old_path = os.path.join(source_data_path, path)
            new_path = os.path.join(target_data_path, path)
            with h5py.File(target_file, mode="a") as h5s:
                if copy_other_data:
                    with h5py.File(source_file, mode="r") as o_h5s:
                        if new_path in h5s:
                            del h5s[new_path]
                        h5s[new_path] = h5py_read_dataset(o_h5s[old_path])
                elif source_file == target_file:
                    h5s[new_path] = h5py.SoftLink(old_path)
                else:
                    relpath = os.path.relpath(source_file, os.path.dirname(target_file))
                    h5s[new_path] = h5py.ExternalLink(relpath, old_path)

    def _patch_nxtomo_flags(self, scan):
        """Insure necessary flags are here"""
        with h5py.File(scan.master_file, mode="a") as h5s:
            instrument_path = os.path.join(scan.entry, "instrument")
            instrument_node = h5s.require_group(instrument_path)
            if "NX_class" not in instrument_node.attrs:
                instrument_node.attrs["NX_class"] = "NXinstrument"
            detector_node = instrument_node.require_group("detector")
            if "NX_class" not in detector_node.attrs:
                detector_node.attrs["NX_class"] = "NXdetector"
            if "data" in instrument_node:
                if "interpretation" not in instrument_node.attrs:
                    instrument_node["data"].attrs["interpretation"] = "image"

            sample_path = os.path.join(scan.entry, "sample")
            sample_node = h5s.require_group(sample_path)
            if "NX_class" not in sample_node:
                sample_node.attrs["NX_class"] = "NXsample"

    def tearDown(self) -> None:
        shutil.rmtree(self.tmpdir)


class TestDaarkFlatPatch(BaseTestAddDarkAndFlats):
    def test(self):
        scan = NXtomoScan(
            self._nx_with_virtual_dataset_with_link.master_file,
            self._nx_with_virtual_dataset_with_link.entry,
        )
        process = DarkFlatPatchTask(
            inputs={
                "data": scan,
                "serialize_output_data": False,
                "configuration": {
                    "darks_start": self.start_dark_url,
                },
            },
        )
        process.definition()
        process.program_version()
        process.program_name()
        # TODO: properties should be part of inputs
        process.run()



from __future__ import annotations

import logging

from ._DataListenerBaseClass import DataListenerBaseClass
from tomwer.core.process.control.datalistener import DataListener
from tomwer.synctools.bliss_sync.blissdata_listener import BlissDataListenerQThread


_logger = logging.getLogger(__name__)


class BlissDataListenerOW(DataListenerBaseClass):
    """
    This widget is used to listen to blissdata at a specific BEACON_HOST
    """

    name = "Blissdata Listener"
    id = "orangecontrib.widgets.tomwer.control.BlissDataListenerOW.BlissDataListenerOW"
    description = (
        "The widget will listen to bliss data in order to retrieve on-going (tomo) bliss-sequence.\n"
        "It must ne started **before** the bliss scan starts.\n"
        "Once the bliss sequence is finished it will be converted to NXtomo(s)"
    )
    icon = "icons/blissdata_listener.svg"
    priority = 10
    keywords = [
        "bliss data",
        "blissdata",
        "tomography",
        "tomwer",
        "listener",
        "datalistener",
        "hdf5",
        "NXtomo",
    ]

    def __init__(self, host_discovery="BEACON_HOST", parent=None):
        super().__init__(host_discovery=host_discovery, parent=parent, uses_rpc=False)
        self._widget.setHostAndPortToolTip(
            "Host and port must be defined from the 'BEACON_HOST' environment before launching the canvas.\nLike for example: `export BEACON_HOST=id00:25000`"
        )

    def delete_listening_thread(self):
        if self._listening_thread is not None:
            self._listening_thread.sigAcquisitionStarted.disconnect(
                self._widget._acquisitionStarted
            )
            self._listening_thread.sigAcquisitionEnded.disconnect(
                self._widget._acquisitionEnded
            )
        DataListener.delete_listening_thread(self)

    def create_listening_thread(self):
        try:
            thread = BlissDataListenerQThread(
                host=self.getHost(), port=self.getPort(), acquisitions=None
            )
        except ConnectionRefusedError:
            _logger.error(self._getErrorMsg(ConnectionRefusedError))
            return None
        except TimeoutError:
            _logger.error(self._getErrorMsg(TimeoutError))
            return None

        # connect thread
        thread.sigAcquisitionStarted.connect(self._widget._acquisitionStarted)
        thread.sigAcquisitionEnded.connect(self._widget._acquisitionEnded)

        return thread

    def _getErrorMsg(self, exception_type):
        return f"Unable to create Blissdata Listener widget. - {exception_type}. Please check that the host and the port are correct from the BEACON_HOST environment variable"



# coding: utf-8
from __future__ import annotations

from tomwer.core.process.reconstruction.nabu.utils import retrieve_lst_of_value_from_str
from tomwer.core.process.reconstruction.scores.params import SABaseParams


class SADeltaBetaParams(SABaseParams):
    """Parameters for the semi-automatic axis calculation"""

    def __init__(self):
        super().__init__()
        self._delta_beta_values = (100,)
        self._selected_db = None

    @property
    def delta_beta_values(self):
        return self._delta_beta_values

    @delta_beta_values.setter
    def delta_beta_values(self, delta_beta_values):
        if isinstance(delta_beta_values, str):
            self._delta_beta_values = retrieve_lst_of_value_from_str(
                delta_beta_values, type_=float
            )
        else:
            self._delta_beta_values = delta_beta_values

    def set_db_selected_value(self, value):
        self._selected_db = value

    @property
    def selected_delta_beta_value(self) -> float | None:
        return self._selected_db

    def to_dict(self) -> dict:
        my_dict = super().to_dict()
        my_dict.update(
            {
                "delta_beta_values": self.delta_beta_values,
                "selected_db_value": self.selected_delta_beta_value,
            }
        )
        return my_dict

    def load_from_dict(self, dict_: dict):
        if not isinstance(dict_, dict):
            raise TypeError(f"dict_ should be an instance of dict not {type(dict_)}")
        super().load_from_dict(dict_)
        if "delta_beta_values" in dict_:
            self.delta_beta_values = dict_["delta_beta_values"]
        if "selected_db_value" in dict_:
            self.set_db_selected_value(value=dict_["selected_db_value"])

    @staticmethod
    def from_dict(dict_):
        params = SADeltaBetaParams()
        params.load_from_dict(dict_=dict_)
        return params

    def check_configuration(self):
        """
        Insure all requested information for processing the SAAXis are here.
        :raises: ValueError if some information are missing
        """
        missing_information = []
        if self.delta_beta_values is None or len(self.delta_beta_values) == 0:
            missing_information.append("no values for center of rotation provided")
        if self.slice_indexes is None:
            missing_information.append("slice index not provided")
        if len(missing_information) > 0:
            missing_information_str = " ; ".join(missing_information)
            raise ValueError(
                f"Some informations are missing: {missing_information_str}"
            )

    def __str__(self):
        return str(self.to_dict())



"""
(Ewoks)Orange Widgets for tomography core processing
"""

NAME = "reconstruction"

ID = "orangecontrib.tomwer.widgets.reconstruction"

DESCRIPTION = "widgets for data reconstruction"

LONG_DESCRIPTION = (
    "Contains widgets for scan reconstruction (mabu slice, nabu volume...)"
)

ICON = "../../widgets/icons/tomwer_build.png"

BACKGROUND = "#C0CCFF"

PRIORITY = 2



from __future__ import annotations

import os
import numpy
import logging
import pathlib
from PIL import Image
from silx.io.url import DataUrl

from .binning import Binning
from .output import OutputFormat

from tomoscan.utils.io import filter_esrf_mounting_points
from tomoscan.esrf.scan.utils import get_data, from_bliss_original_file_to_raw

from tomwer.core.process.drac.output import DATASET_GALLERY_DIR_NAME
from tomwer.core.tomwer_object import TomwerObject
from tomwer.core.scan.scanbase import TomwerScanBase


_logger = logging.getLogger(__name__)

__all__ = [
    "DracDatasetBase",
]


class DracDatasetBase:
    """
    Abstract class for a drac dataset (that should be publish to the data portal).
    A drac dataset is defined by the following elements:
    * a directory
    * metadata
    """

    def __init__(self, data_dir: str, tomo_obj: TomwerObject) -> None:
        """
        :param data_dir: directory containing processed data
        :param tomo_obj: tomwer object referee for the processing
        """
        data_dir = pathlib.Path(data_dir).resolve()
        self._data_dir = filter_esrf_mounting_points(str(data_dir))
        self._tomo_obj = tomo_obj
        self._gallery_file_binning = Binning.FOUR_BY_FOUR
        self._gallery_output_format = OutputFormat.PNG
        self._gallery_overwrite = True

    @property
    def gallery_output_binning(self) -> Binning:
        "Binning factor to be used when saving gallery images"
        return self._gallery_file_binning

    @gallery_output_binning.setter
    def gallery_output_binning(self, binning: Binning):
        binning = Binning(binning)
        self._gallery_file_binning = binning

    @property
    def gallery_output_format(self) -> OutputFormat:
        "Output format of the images stored in the gallery"
        return self._gallery_output_format

    @gallery_output_format.setter
    def gallery_output_format(self, format: OutputFormat):
        format = OutputFormat(format)
        self._gallery_output_format = format

    @property
    def gallery_overwrite(self) -> bool:
        return self._gallery_overwrite

    @gallery_overwrite.setter
    def gallery_overwrite(self, overwrite: bool) -> None:
        self._gallery_overwrite = overwrite

    @property
    def dataset_name(self) -> str:
        """
        Return the name of the dataset to be given to the drac dataset.
        A drac dataset is store according to a (dataset_path, dataset_name). By default if several are pushed with the same key,
        only the most recent one will be stored
        """
        raise NotImplementedError("Base class")

    def get_gallery_dir(self) -> str:
        return os.path.join(self.data_dir, DATASET_GALLERY_DIR_NAME)

    def build_gallery(self):
        """callback to build the drac-dataset gallery"""
        pass

    @property
    def data_dir(self):
        return self._data_dir

    @property
    def tomo_obj(self):
        return self._tomo_obj

    @property
    def metadata(self) -> dict:
        raise NotImplementedError("Base class")

    @property
    def bliss_raw_datasets(self) -> tuple[str]:
        raise NotImplementedError("Base class")

    @staticmethod
    def from_scan_to_raws(scan: TomwerScanBase) -> tuple[str]:
        assert isinstance(
            scan, TomwerScanBase
        ), f"scan should be an instance of {TomwerScanBase}. Got {type(scan)}"
        bliss_original_files = scan.get_bliss_original_files()
        if bliss_original_files is None:
            bliss_original_files = ()
        return tuple(
            [
                from_bliss_original_file_to_raw(bliss_original_file)
                for bliss_original_file in bliss_original_files
            ]
        )

    def to_dict(self):
        return {
            "output_dir": self.data_dir,
            "drac_metadata": self.metadata,
        }

    def save_to_gallery(
        self,
        output_file_name: str,
        image: numpy.ndarray | DataUrl,
        lower_bound: float | None = None,
        upper_bound: float | None = None,
    ) -> None:
        format = self.gallery_output_format
        overwrite = self.gallery_overwrite
        binning = self.gallery_output_binning

        gallery_dir = self.get_gallery_dir()
        os.makedirs(gallery_dir, exist_ok=True)

        if isinstance(image, DataUrl):
            image = get_data(image)

        if image.ndim == 3 and image.shape[0] == 1:
            image = image.reshape(image.shape[1:])
        elif image.ndim != 2:
            raise ValueError(f"Only 2D grayscale images are handled. Got {image.shape}")

        # If both bounds are provided, apply clamping and normalization
        # If not it will use the min/max of each image
        if lower_bound is not None and upper_bound is not None:
            image = numpy.clip(image, lower_bound, upper_bound)
            image = image - lower_bound
            image *= 255.0 / (upper_bound - lower_bound)
        else:
            min_val, max_val = image.min(), image.max()
            image = image - min_val
            if max_val > min_val:  # Avoid division by zero if image is constant
                image *= 255.0 / (max_val - min_val)

        # Binning the data if required
        image = Binning._bin_data(data=image, binning=binning)

        # Convert to a PIL Image and save
        img = Image.fromarray(image.astype(numpy.uint8), mode="L")
        if not output_file_name.endswith(f".{format.value}"):
            output_file_name = output_file_name + f".{format.value}"
        if not overwrite and os.path.exists(output_file_name):
            raise OSError(f"File already exists ({output_file_name})")
        img.save(output_file_name)



# coding: utf-8
from __future__ import annotations

import logging
import operator
import os

logger = logging.getLogger(__name__)


def get_vol_file_shape(file_path):
    ddict = {}
    f = open(file_path, "r")
    lines = f.readlines()
    for line in lines:
        if "=" not in line:
            continue
        line_str = line.rstrip().replace(" ", "")
        line_str = line_str.split("#")[0]
        key, value = line_str.split("=")
        ddict[key.lower()] = value

    dimX = None
    dimY = None
    dimZ = None

    if "num_z" in ddict:
        dimZ = int(ddict["num_z"])
    if "num_y" in ddict:
        dimY = int(ddict["num_y"])
    if "num_x" in ddict:
        dimX = int(ddict["num_x"])

    return (dimZ, dimY, dimX)


def orderFileByLastLastModification(folder, fileList):
    """Return the list of files sorted by time of last modification.
    From the oldest to the newest modify"""
    res = {}
    for f in fileList:
        res[os.path.getmtime(os.path.join(folder, f))] = f

    return sorted(res.items(), key=operator.itemgetter(0))



# import pickle

# from orangecanvas.scheme.readwrite import literal_dumps

# from tomwer.tests.conftest import qtapp  # noqa F401

# from orangecontrib.tomwer.widgets.reconstruction.NabuHelicalPrepareWeightsDoubleOW import (
#     NabuHelicalPrepareWeightsDoubleOW,
# )


# def test_NabuHelicalPrepareWeightsDoubleOW(
#     qtapp,  # noqa F811
# ):
#     """simple test of the _DeltaBetaSelectorDialog"""
#     widget = NabuHelicalPrepareWeightsDoubleOW()
#     widget._loadSettings()
#     # test settings serialization
#     pickle.dumps(widget._ewoks_default_inputs)
#     literal_dumps(widget._ewoks_default_inputs)



import pytest
from ewoksorange.bindings import owsconvert
from tomwer.tests.datasets import TomwerCIDatasets


ows_files_to_convert = ("from_nxtomo_to_volume.ows",)


@pytest.mark.parametrize("orange_file_name", ows_files_to_convert)
def test_conversion(orange_file_name):
    """test orange file can be converted to ewoks Graph"""
    scheme_file = TomwerCIDatasets.get_dataset(f"workflows/{orange_file_name}")
    graph = owsconvert.ows_to_ewoks(scheme_file)
    # make sure settings where correctly saved
    assert len(graph.graph.nodes) == 7
    assert len(graph.graph.edges) == 6
    # make sure the settings are properly loaded
    for node in graph.graph.nodes.values():
        assert node["task_type"] == "class"
        assert node["task_identifier"].startswith("tomwer")
    # check settings of some specific nodes
    # dark and flat
    dark_flat_node = graph.graph.nodes["1"]
    assert (
        dark_flat_node["task_identifier"]
        == "tomwer.core.process.reconstruction.darkref.darkrefscopy.DarkRefsCopy"
    )
    assert "default_inputs" in dark_flat_node

    def fileter_dark_ref_params(my_dict):
        return my_dict["name"] == "dark_ref_params"

    dark_ref_params_settings = tuple(
        filter(fileter_dark_ref_params, dark_flat_node["default_inputs"])
    )
    assert len(dark_ref_params_settings) == 1
    dark_ref_params_as_dict = dark_ref_params_settings[0]["value"]

    assert "DOWHEN" in dark_ref_params_as_dict
    assert "DARKCAL" in dark_ref_params_as_dict
    assert "DKFILE" in dark_ref_params_as_dict

    # axis
    axis_node = graph.graph.nodes["2"]
    assert (
        axis_node["task_identifier"]
        == "tomwer.core.process.reconstruction.axis.axis.AxisTask"
    )
    assert "default_inputs" in axis_node
    assert len(axis_node["default_inputs"]) == 2  # axis_params + gui metadata

    def filter_axis_params(my_dict):
        return my_dict["name"] == "axis_params"

    axis_params_settings = tuple(
        filter(filter_axis_params, axis_node["default_inputs"])
    )
    assert len(axis_params_settings) == 1
    axis_params_as_dict = axis_params_settings[0]["value"]

    assert "MODE" in axis_params_as_dict
    assert "POSITION_VALUE" in axis_params_as_dict
    assert "SINOGRAM_LINE" in axis_params_as_dict

    # nabu slice
    nabu_slice_node = graph.graph.nodes["3"]
    assert (
        nabu_slice_node["task_identifier"]
        == "tomwer.core.process.reconstruction.nabu.nabuslices.NabuSlicesTask"
    )
    assert "default_inputs" in axis_node

    def filter_nabu_params(my_dict):
        return my_dict["name"] == "nabu_params"

    nabu_params_settings = tuple(
        filter(filter_nabu_params, nabu_slice_node["default_inputs"])
    )
    assert len(nabu_params_settings) == 1
    nabu_params_as_dict = nabu_params_settings[0]["value"]

    assert "preproc" in nabu_params_as_dict
    assert "phase" in nabu_params_as_dict
    assert "tomwer_slices" in nabu_params_as_dict

    # nabu volume
    nabu_volume_node = graph.graph.nodes["4"]
    assert (
        nabu_volume_node["task_identifier"]
        == "tomwer.core.process.reconstruction.nabu.nabuvolume.NabuVolumeTask"
    )
    assert "default_inputs" in axis_node

    def filter_nabu_volume_params(my_dict):
        return my_dict["name"] == "nabu_volume_params"

    nabu_volume_params_settings = tuple(
        filter(filter_nabu_volume_params, nabu_volume_node["default_inputs"])
    )
    assert len(nabu_volume_params_settings) == 1
    nabu_volume_params_as_dict = nabu_volume_params_settings[0]["value"]

    assert "start_z" in nabu_volume_params_as_dict
    assert "end_z" in nabu_volume_params_as_dict



from __future__ import annotations

import io
import json
import logging
import os
from glob import glob
import functools
import pathlib
from typing import Iterable

import numpy
from silx.io.url import DataUrl
from enum import Enum as _Enum
from pathlib import Path

from tomoscan.identifier import VolumeIdentifier
from tomoscan.normalization import IntensityNormalization
from tomoscan.volumebase import VolumeBase
from tomoscan.identifier import BaseIdentifier
from tomoscan.esrf.volume.utils import guess_volumes
from tomoscan.utils.io import filter_esrf_mounting_points

from tomwer.core.tomwer_object import TomwerObject
from tomwer.core.utils.ftseriesutils import orderFileByLastLastModification
from tomwer.io.utils.raw_and_processed_data import (
    to_raw_data_path,
    to_processed_data_path,
)

logger = logging.getLogger(__name__)


class TomwerScanBase(TomwerObject):
    """
    Interface to extend tomoscan.TomoScanBase with custom functionality.
    ..warning:: All classes inheriting from this interface must also inherit from a child class of tomoscan.TomoScanBase
    """

    _DICT_DARK_REF_KEYS = "dark_ref_params"

    _DICT_NABU_RP_KEY = "nabu_params"

    _DICT_AXIS_KEYS = "axis_params"

    _DICT_SA_AXIS_KEYS = "sa_axis_params"

    _DICT_SA_DELTA_BETA_KEYS = "sa_delta_beta_params"

    _DICT_NORMALIZATION_KEY = "norm_params"

    VALID_RECONS_EXTENSION = ".edf", ".npy", ".npz", ".hdf5", ".tiff", ".jp2", ".vol"

    def __init__(self, overwrite_proc_file=False):
        super().__init__()
        self._stitching_metadata = None
        self._reconstructions = []

        self._nabu_params = None
        """nabu reconstruction parameters"""
        self._axis_params = None
        """Axis parameters"""
        self._saaxis_params = None
        """Information relative to saaxis"""
        self._sa_delta_beta_params = None
        """Information regarding sa_delta_beta_params"""
        self._dark_ref_params = None
        """Information regarding dark - ref reconstruction"""
        self._cache_proj_urls = None
        """cache for the projection urls"""
        self._cache_radio_axis = {}
        """cache for the radio axis. Key is tuple (mode, nearest), value is
        (url1, url2)"""
        self._notify_ffc_rsc_missing = True
        """Should we notify the user if ffc fails because cannot find dark or
        flat. Used to avoid several warnings. Only display one"""
        self._latest_reconstructions = []
        "list of url related to latest slice reconstruction from nabu"
        self._latest_vol_reconstructions = []
        """list of url related to latest volume reconstruction from nabu"""
        self._reconstruction_paths = set()
        """set of paths containing reconstructions of the scan. Make sure we have an entry point for it"""
        self._proposal_name = None

    def _clear_heavy_cache(self):
        """For scan for now we don't want to remove any information from the cache.
        Mor eusefull for volume use case
        """
        pass

    @property
    def resolved_path(self) -> str | None:
        """
        return tomoscan 'path' with link resolved
        """
        if self.path is None:  # pylint: disable=E1101
            return None
        return filter_esrf_mounting_points(
            str(Path(self.path).resolve())  # pylint: disable=E1101
        )

    @property
    def working_directory(self) -> Path | None:
        """
        working directory once all links have been solved (required for submitting jobs to slurm)
        """
        if self.working_directory is None:
            return None
        return pathlib.Path(self.working_directory)

    def clear_cache(self):
        self._cache_proj_urls = None
        self._notify_ffc_rsc_missing = True
        super().clear_cache()

    def _flat_field_correction(
        self,
        data,
        index_proj: int | None,
        dark,
        flat1,
        flat2,
        index_flat1: int,
        index_flat2: int,
    ):
        """
        compute flat field correction for a provided data from is index
        one dark and two flats (require also indexes)
        """
        if not isinstance(data, numpy.ndarray):
            raise TypeError("data is expected to be an instance of numpy.ndarray")
        can_process = True

        if dark is None:
            if self._notify_ffc_rsc_missing:
                logger.warning("cannot make flat field correction, dark not found")
            can_process = False

        if dark is not None and dark.ndim != 2:
            logger.error(
                "cannot make flat field correction, dark should be of " "dimension 2"
            )
            can_process = False

        if flat1 is None:
            if self._notify_ffc_rsc_missing:
                logger.warning("cannot make flat field correction, flat not found")
            can_process = False
        elif not isinstance(flat1, numpy.ndarray):
            raise TypeError("flat1 is expected to be an instance of numpy.ndarray")
        else:
            if flat1.ndim != 2:
                logger.error(
                    "cannot make flat field correction, flat should be of "
                    "dimension 2"
                )
                can_process = False
            if flat2 is not None and flat1.shape != flat2.shape:
                logger.error("the tow flats provided have different shapes.")
                can_process = False

        if dark is not None and flat1 is not None and dark.shape != flat1.shape:
            logger.error("Given dark and flat have incoherent dimension")
            can_process = False

        if dark is not None and data.shape != dark.shape:
            logger.error(
                "Image has invalid shape. Cannot apply flat field" "correction it"
            )
            can_process = False

        if can_process is False:
            self._notify_ffc_rsc_missing = False
            return data

        if flat2 is None:
            flat_value = flat1
        else:
            # compute weight and clip it if necessary
            if index_proj is None:
                w = 0.5
            else:
                w = (index_proj - index_flat1) / (index_flat2 - index_flat1)
                w = min(1, w)
                w = max(0, w)
            flat_value = flat1 * w + flat2 * (1 - w)

        div = flat_value - dark
        div[div == 0] = 1
        return (data - dark) / div

    @property
    def reconstructions(self):
        """list of reconstruction files"""
        return self._reconstructions

    @reconstructions.setter
    def reconstructions(self, reconstructions):
        self._reconstructions = reconstructions

    @property
    def reconstruction_paths(self):
        return self._reconstruction_paths

    def add_reconstruction_path(self, path: str):
        self._reconstruction_paths.add(path)

    @property
    def nabu_recons_params(self):
        return self._nabu_params

    @nabu_recons_params.setter
    def nabu_recons_params(self, recons_params):
        self._nabu_params = recons_params

    @property
    def axis_params(self):
        return self._axis_params

    @axis_params.setter
    def axis_params(self, parameters):
        self._axis_params = parameters

    @property
    def saaxis_params(self):
        return self._saaxis_params

    @saaxis_params.setter
    def saaxis_params(self, saaxis_params):
        self._saaxis_params = saaxis_params

    @property
    def sa_delta_beta_params(self):
        return self._sa_delta_beta_params

    @sa_delta_beta_params.setter
    def sa_delta_beta_params(self, sa_delta_beta_params):
        self._sa_delta_beta_params = sa_delta_beta_params

    # TODO: change name. Should be generalized to return Dataurl
    def getReconstructedFilesFromParFile(self, with_index):
        raise NotImplementedError("Base class")

    def projections_with_angle(self):
        raise NotImplementedError("Base class")

    def scan_dir_name(self) -> str | None:
        """return name of the directory containing the acquisition"""
        raise NotImplementedError("Base class")

    def scan_basename(self) -> str | None:
        """return basename of the directory containing the acquisition"""
        raise NotImplementedError("Base class")

    def scan_parent_dir_basename(self) -> str | None:
        """return parent basename of the directory containing the acquisition"""
        raise NotImplementedError("Base class")

    @functools.lru_cache(maxsize=3)
    def get_opposite_projections(self, mode) -> tuple:
        """
        Return the two best opposite projections for the required mode.
        """
        radios_with_angle = self.projections_with_angle()
        angles = numpy.array(
            tuple(
                filter(
                    lambda a: numpy.issubdtype(type(a), numpy.number),
                    radios_with_angle.keys(),
                )
            )
        )

        from tomwer.core.process.reconstruction.axis.anglemode import CorAngleMode
        from ..process.reconstruction.axis.params import (
            AxisResource,
        )  # avoid cyclic import

        mode = CorAngleMode.from_value(mode)

        if len(angles) < 2:
            logger.warning("less than two angles found. Unable to get opposite angles")
            return None, None

        initial_angle = angles[0]
        if mode in (CorAngleMode.use_0_180, CorAngleMode.manual_selection):
            couples = (initial_angle, initial_angle + 180.0)
        elif mode is CorAngleMode.use_90_270:
            couples = (initial_angle + 90.0, initial_angle + 270.0)
        else:
            raise ValueError(f"{mode} is not handle")

        def find_nearest(angles: numpy.ndarray, angle: float):
            if len(angles) == 0:
                return None
            dist = numpy.abs(angles - angle)
            idx = dist.argmin()
            if isinstance(idx, numpy.ndarray):
                idx = idx[0]
            return angles[idx]

        nearest_c1 = find_nearest(angles=angles, angle=couples[0])
        nearest_c2 = find_nearest(angles=angles, angle=couples[1])
        if nearest_c1 is not None and nearest_c2 is not None:
            return (
                AxisResource(radios_with_angle[nearest_c1], angle=nearest_c1),
                AxisResource(radios_with_angle[nearest_c2], angle=nearest_c2),
            )
        else:
            return None, None

    def data_flat_field_correction(
        self, data: numpy.ndarray, index: int | None = None
    ) -> numpy.ndarray:
        """Apply flat field correction on the given data

        :param data: the data to apply correction on
        :param index: index of the data in the acquisition
                                      sequence
        :return: corrected data
        """
        raise NotImplementedError("Base class")

    def to_dict(self):
        res = {}
        # nabu reconstruction parameters
        if self._nabu_params:
            res[self._DICT_NABU_RP_KEY] = (
                self.nabu_recons_params
                if isinstance(self.nabu_recons_params, dict)
                else self.nabu_recons_params.to_dict()
            )
        else:
            res[self._DICT_NABU_RP_KEY] = None
        # axis reconstruction parameters
        if self.axis_params is None:
            res[self._DICT_AXIS_KEYS] = None
        else:
            res[self._DICT_AXIS_KEYS] = self.axis_params.to_dict()
        # saaxis reconstruction parameters
        if self.saaxis_params is None:
            res[self._DICT_SA_AXIS_KEYS] = None
        else:
            res[self._DICT_SA_AXIS_KEYS] = self.saaxis_params.to_dict()
        # sa delta-beta reconstruction parameters
        if self._sa_delta_beta_params is None:
            res[self._DICT_SA_DELTA_BETA_KEYS] = None
        else:
            res[self._DICT_SA_DELTA_BETA_KEYS] = self.sa_delta_beta_params.to_dict()
        # dark ref
        if self._dark_ref_params is None:
            res[self._DICT_DARK_REF_KEYS] = None
        else:
            res[self._DICT_DARK_REF_KEYS] = self._dark_ref_params.to_dict()
        # normalization
        if self.intensity_normalization is None:
            res[self._DICT_NORMALIZATION_KEY] = None
        else:
            res[self._DICT_NORMALIZATION_KEY] = self.intensity_normalization.to_dict()

        return res

    def load_from_dict(self, desc):
        from tomwer.core.process.reconstruction.axis.params import (
            AxisRP,
        )  # avoid cyclic import

        if isinstance(desc, io.TextIOWrapper):
            data = json.load(desc)
        else:
            data = desc
        if not (
            self.DICT_PATH_KEY in data  # pylint: disable=E1101
            and data[self.DICT_TYPE_KEY] == self._TYPE  # pylint: disable=E1101
        ):
            raise ValueError("Description is not an EDFScan json description")

        assert self.DICT_PATH_KEY in data  # pylint: disable=E1101
        # load axis reconstruction parameters
        axis_params = data.get(self._DICT_AXIS_KEYS, None)
        if axis_params is not None:
            self.axis_params = AxisRP.from_dict(axis_params)
        # load nabu reconstruction parameters
        if self._DICT_NABU_RP_KEY in data:
            self._nabu_params = data[self._DICT_NABU_RP_KEY]
        # load dark-ref parameters
        dark_ref_params = data.get(self._DICT_DARK_REF_KEYS, None)
        if dark_ref_params is not None:
            from tomwer.core.process.reconstruction.darkref.params import DKRFRP

            self._dark_ref_params = DKRFRP.from_dict(dark_ref_params)
        # load normalization
        intensity_normalization = data.get(self._DICT_NORMALIZATION_KEY, None)
        if intensity_normalization is not None:
            self.intensity_normalization = IntensityNormalization.from_dict(
                intensity_normalization
            )
        # load saaxis parameters
        saaxis_params = data.get(self._DICT_SA_AXIS_KEYS, None)
        if saaxis_params is not None:
            from tomwer.core.process.reconstruction.saaxis.params import SAAxisParams

            self._saaxis_params = SAAxisParams.from_dict(saaxis_params)
        # load sa delta beta parameters
        sa_delta_beta_params = data.get(self._DICT_SA_DELTA_BETA_KEYS, None)
        if sa_delta_beta_params is not None:
            from tomwer.core.process.reconstruction.sadeltabeta.params import (
                SADeltaBetaParams,
            )

            self._sa_delta_beta_params = SADeltaBetaParams.from_dict(
                sa_delta_beta_params
            )

    def equal(self, other):
        """

        :param other: instance to compare with
        :return: True if instance are equivalent

        :note: we cannot use the __eq__ function because this object need to be picklable
        """
        return (
            isinstance(other, self.__class__)
            or isinstance(self, other.__class__)
            and self.type == other.type  # pylint: disable=E1101
            and self.nabu_recons_params == other.nabu_recons_params
            and self.path == other.path  # pylint: disable=E1101
        )

    def get_sinogram(
        self, line: int, subsampling: int = 1, norm_method=None, **kwargs
    ) -> numpy.ndarray:
        """
        extract the sinogram from projections

        :param line: which sinogram we want
        :param subsampling: subsampling to apply on the sinogram
        :return: sinogram from the radio lines
        """
        raise NotImplementedError("Base class")

    def get_normed_sinogram(self, line: int, subsampling: int = 1) -> numpy.ndarray:
        """
        Util to get the sinogram normed with settings currently defined
        on the 'intensity_normalization' property

        :param line:
        :param subsampling:
        :return:
        """
        return self.get_sinogram(
            line=line,
            subsampling=subsampling,
            norm_method=self.intensity_normalization.method,
            **self.intensity_normalization.get_extra_infos(),
        )

    def __str__(self):
        raise NotImplementedError("Base class")

    @staticmethod
    def get_pyhst_recons_file(scanID):
        """Return the .par file used for the current reconstruction if any.
        Otherwise return None"""
        if scanID == "":
            return None

        if scanID is None:
            raise RuntimeError("No current acquisition to validate")
        assert type(scanID) is str
        assert os.path.isdir(scanID)
        folderID = os.path.basename(scanID)
        # look for fasttomo files ending by slice.par
        parFiles = glob(os.path.join(scanID + folderID) + "*_slice.par")
        if len(parFiles) > 0:
            return orderFileByLastLastModification(scanID, parFiles)[-1]
        else:
            return None

    def _deduce_transfert_scan(self, output_dir):
        """
        Create the scan that will be generated if this scan is
        copy to the output_dir

        :param output_dir:
        """
        raise NotImplementedError("Base class")

    def get_proj_angle_url(
        self, use_cache: bool = True, *args, **kwargs
    ) -> dict[int | str, DataUrl]:
        """
        retrieve the url for each projections (including the alignment /
        return one) and associate to each (if possible) the angle.
        Alignment angle are set as angle (1) to specify that this is an
        alignment one.
        :param use_cache:
        :return: dictionary with angle (str or int) as key and url as value
        """
        raise NotImplementedError("Base class")

    def get_reconstructed_slices(self) -> set[str]:
        """
        Return a set of all the reconstructed **slices** that can be discovered for the scan (it will only look for reconstructed slices and not the volumes!!!!)
        slices are given as a string representing volume identifier
        """
        from tomwer.core.process.reconstruction.output import (
            PROCESS_FOLDER_RECONSTRUCTED_SLICES,
        )  # avoid cyclic import

        all_recons_ids = set()
        recons_paths = self.reconstruction_paths
        # extend set reconstruction paths to some 'default' paths.
        if self.working_directory is not None:
            # default output folder if using working directory
            working_dir_output_folder = os.path.join(
                str(self.working_directory),
                PROCESS_FOLDER_RECONSTRUCTED_SLICES,
            )
            # add possible output folder: in raw_data (deprecated)
            raw_data_output_folder = to_raw_data_path(working_dir_output_folder)
            recons_paths.add(raw_data_output_folder)
            # add expected output folder: in processed data
            processed_data_output_folder = to_processed_data_path(
                working_dir_output_folder
            )
            recons_paths.add(processed_data_output_folder)

        for path in recons_paths:
            all_recons_ids.update(self.get_reconstructed_slices_from_path(path))
        return all_recons_ids

    def get_reconstructed_slices_from_path(self, path, check_url=False) -> set[str]:
        """
        Look for some reconstructed slices at a specific position
        slices are given as a string representing volume identifier
        """
        if path is None or not os.path.exists(path):
            return set()

        volumes = guess_volumes(path) or ()
        # note: guess_volumes will find a volume only if the path fits a volume location
        # like a hdf5 file containing a volume or a top level EDF volume folder.
        # In the case path = xxx/reconstructed_slices for example it returns an empty set.
        if len(volumes) > 0:
            return set(
                [volume.get_identifier().to_str() for volume in guess_volumes(path)]
            )
        elif os.path.isdir(path):
            # in the case the location is a folder and doesn't fit any volume
            # we want to look into sub folders
            from tomwer.core.process.reconstruction.saaxis.saaxis import (
                DEFAULT_RECONS_FOLDER as MULTI_COR_DEFAULT_FOLDER,
            )
            from tomwer.core.process.reconstruction.sadeltabeta.sadeltabeta import (
                DEFAULT_RECONS_FOLDER as MULTI_DB_DEFAULT_FOLDER,
            )
            from tomwer.core.process.reconstruction.nabu.settings import (
                NABU_CFG_FILE_FOLDER,
            )

            res = set()
            filtered_name = (
                "steps_file_basename_nabu_sinogram_save_step",
                MULTI_COR_DEFAULT_FOLDER,
                MULTI_DB_DEFAULT_FOLDER,
                NABU_CFG_FILE_FOLDER,
            )
            # filter files that we already know they won't contain any reconstructed volume
            sub_files = filter(
                lambda name: name not in filtered_name,
                os.listdir(path),
            )
            for f in sub_files:
                current_path = os.path.join(path, f)
                res.update(
                    self.get_reconstructed_slices_from_path(
                        path=current_path, check_url=check_url
                    )
                )
            return res
        return set()

    @property
    def latest_reconstructions(self):
        """List of latest slices reconstructions (as VolumeIdentifier) - single slice volume"""
        return self._latest_reconstructions

    @property
    def latest_vol_reconstructions(self):
        """List of latest volume reconstructions (as VolumeIdentifier)"""
        return self._latest_vol_reconstructions

    def clear_latest_reconstructions(self):
        self._latest_reconstructions = []

    def set_latest_reconstructions(self, urls: Iterable):
        if urls is None:
            self._latest_reconstructions = None
        else:
            self._latest_reconstructions = list(
                [self._process_volume_url(url) for url in urls]
            )

    def add_latest_reconstructions(self, urls: Iterable):
        self._latest_reconstructions.extend(urls)

    def clear_latest_vol_reconstructions(self):
        self._latest_vol_reconstructions = []

    @staticmethod
    def _process_volume_url(url):
        if isinstance(url, str):
            return VolumeIdentifier.from_str(url)
        elif isinstance(url, VolumeIdentifier):
            return url
        elif isinstance(url, VolumeBase):
            return url.get_identifier()
        else:
            raise TypeError(
                f"url should be a {VolumeIdentifier} or a string reprenseting a {VolumeIdentifier}. {type(url)} provided instead"
            )

    def set_latest_vol_reconstructions(self, volume_identifiers: Iterable | None):
        volume_identifiers = volume_identifiers or tuple()
        self._latest_vol_reconstructions = list(
            [self._process_volume_url(url) for url in volume_identifiers]
        )

    def add_latest_vol_reconstructions(self, volume_identifiers: tuple):
        assert isinstance(
            volume_identifiers, tuple
        ), "volume_identifiers is expected to be a tuple"
        self._latest_vol_reconstructions.extend(
            self._process_volume_url(volume_identifier)
            for volume_identifier in volume_identifiers
        )

    def _update_latest_recons_identifiers(self, old_path, new_path):
        def update_identifier(identifier):
            assert isinstance(
                identifier, BaseIdentifier
            ), f"identifier is expected to be an instance of {BaseIdentifier}"
            # small hack as this is not much used: simply try to replace a path by another. this is only used by the data transfer and EDF / SPEC
            # this time is almost over
            # FIXME: correct way to do this would be to recreate the volume, modify file or folder path and
            # recreate the new identifier
            identifier.replace(old_path, new_path, 1)

        self._latest_reconstructions = [
            update_identifier(identifier=identifier)
            for identifier in self._latest_reconstructions
        ]

        self._latest_vol_reconstructions = [
            update_identifier(identifier=identifier)
            for identifier in self._latest_vol_reconstructions
        ]

    def get_url_proj_index(self, url):
        """Return the index in the acquisition from the url"""

        def invert_dict(ddict):
            res = {}
            if ddict is not None:
                for key, value in ddict.items():
                    assert isinstance(value, DataUrl)
                    res[value.path()] = key
            return res

        proj_inv_url_to_index = invert_dict(self.projections)  # pylint: disable=E1101
        alig_inv_url_to_index = invert_dict(
            self.alignment_projections  # pylint: disable=E1101
        )
        if url.path() in proj_inv_url_to_index:
            return proj_inv_url_to_index[url.path()]
        elif url.path() in alig_inv_url_to_index:
            return alig_inv_url_to_index[url.path()]
        else:
            return None

    def get_nabu_dataset_info(self, binning=1, binning_z=1, proj_subsampling=1) -> dict:
        """

        :return: nabu dataset descriptor
        """
        raise NotImplementedError("Base class")

    def to_nabu_dataset_analyser(self):
        """Return the equivalent DatasetAnalyzer for nabu"""
        raise NotImplementedError("Base class")

    def get_proposal_name(self) -> str | None:
        return self._proposal_name

    def set_proposal_name(self, proposal_name: str) -> None:
        self._proposal_name = proposal_name


class _TomwerBaseDock(object):
    """
    Internal class to make difference between a simple TomoBase output and
    an output for a different processing (like scanvalidator.UpdateReconsParam)
    """

    def __init__(self, tomo_instance):
        self.__instance = tomo_instance

    @property
    def instance(self):
        return self.__instance


def _containsDigits(input_):
    return any(char.isdigit() for char in input_)


class NormMethod(_Enum):
    MANUAL_ROI = "manual ROI"
    AUTO_ROI = "automatic ROI"
    DATASET = "from dataset"



import contextlib
import logging
import os
import shutil
import tempfile
import unittest

import fabio.edfimage
import numpy

from tomwer.core.process.reconstruction.nabu import nabuslices
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils.scanutils import MockEDF
from tomwer.tests.datasets import TomwerCIDatasets

logging.disable(logging.INFO)


class TestScanFactory(unittest.TestCase):
    """Make sure the Scan factory is correctly working. Able to detect the valid
    scan type for a given file / directory
    """

    def test_no_scan(self):
        scan_dir = tempfile.mkdtemp()
        with self.assertRaises(ValueError):
            ScanFactory.create_scan_object(scan_dir)

    def test_scan_edf(self):
        scan_dir = TomwerCIDatasets.get_dataset(
            "edf_datasets/test10",
        )
        scan = ScanFactory.create_scan_object(scan_dir)
        self.assertTrue(isinstance(scan, EDFTomoScan))


class TestScanValidatorFindNabuFiles(unittest.TestCase):
    """
    Make sure files produced by nabu can be seen as reconstruction files
    """

    class FileCreator(contextlib.AbstractContextManager):
        def __init__(self, file_path):
            self.__file_path = file_path

        def __enter__(self):
            with open(self.__file_path, "w"):
                pass

        def __exit__(self, exc_type, exc_val, exc_tb):
            if os.path.exists(self.__file_path):
                os.remove(self.__file_path)

    def setUp(self) -> None:
        self.scan = MockEDF.mockScan(tempfile.mkdtemp(), nPagRecons=0, nRecons=0)
        assert (
            len(self.scan.get_reconstructions_paths(self.scan.path)) == 0
        ), "make sure the initial environment is clear"

    def tearDown(self) -> None:
        shutil.rmtree(self.scan.path)

    def test(self) -> None:
        for slice_index in (None, 1, 15):
            for delta_beta in (None, 100, 500):
                for pag in (True, False):
                    for ctf in (True, False):
                        file_name = nabuslices.SingleSliceRunner.get_file_basename_reconstruction(
                            slice_index=slice_index,
                            scan=self.scan,
                            pag=pag if delta_beta is not None else False,
                            ctf=ctf if delta_beta is not None else False,
                            db=delta_beta,
                            axis="XY",
                        )
                        with self.subTest(
                            slice_index=slice_index,
                            pag_db=delta_beta,
                        ):
                            file_path = os.path.join(
                                self.scan.path, file_name + ".hdf5"
                            )

                            # simple context Manager which create the file
                            # and remove it when leave
                            with TestScanValidatorFindNabuFiles.FileCreator(file_path):
                                n_to_discover = 1
                                if slice_index is None:
                                    n_to_discover = 0
                                self.assertEqual(
                                    len(
                                        self.scan.get_reconstructions_paths(
                                            self.scan.path
                                        )
                                    ),
                                    n_to_discover,
                                )


class TestScanValidatorFindPyHSTFiles(unittest.TestCase):
    """Function testing the getReconstructionsPaths function is correctly
    functioning"""

    DIM_MOCK_SCAN = 10

    N_RADIO = 20
    N_RECONS = 10
    N_PAG_RECONS = 5

    def setUp(self):
        # create scan folder
        self.path = tempfile.mkdtemp()
        MockEDF.mockScan(
            scanID=self.path,
            nRadio=self.N_RADIO,
            nRecons=self.N_RECONS,
            nPagRecons=self.N_PAG_RECONS,
            dim=self.DIM_MOCK_SCAN,
        )
        basename = os.path.basename(self.path)

        # add some random files
        for _file in ("45gfdgfg1.edf", "465slicetest1.edf", "slice_ab.edf"):
            with open(os.path.join(self.path, basename + _file), "w+") as ofile:
                ofile.write("test")

    def tearDown(self):
        if os.path.isdir(self.path):
            shutil.rmtree(self.path)

    def testGetRadioPaths(self):
        nFound = len(EDFTomoScan.get_proj_urls(self.path))
        self.assertTrue(nFound == self.N_RADIO)

    def testGetReconstructionsPaths(self):
        reconstruction_found = EDFTomoScan.get_reconstructions_paths(self.path)
        self.assertEqual(len(reconstruction_found), self.N_RECONS + self.N_PAG_RECONS)

    def testGetSliceReconstruction(self):
        self.assertTrue(
            EDFTomoScan.get_index_reconstructed("dfadf_slice_32.edf", "dfadf") == 32
        )
        self.assertTrue(
            EDFTomoScan.get_index_reconstructed("dfadf_slice_slice_002.edf", "dfadf")
            == 2
        )
        self.assertTrue(
            EDFTomoScan.get_index_reconstructed("scan_3slice_0050.edf", "scan_3") == 50
        )
        self.assertTrue(
            EDFTomoScan.get_index_reconstructed("scan3slice_012.edf", "scan3") == 12
        )


class TestRadioPath(unittest.TestCase):
    """Test static method getRadioPaths for EDFTomoScan"""

    def test(self):
        files = [
            "essai1_0008.edf",
            "essai1_0019.edf",
            "essai1_0030.edf",
            "essai1_0041.edf",
            "essai1_0052.edf",
            "essai1_0063.edf",
            "essai1_0074.edf",
            "essai1_0085.edf",
            "essai1_0096.edf",
            "essai1_.par",
            "refHST0100.edf",
            "darkend0000.edf",
            "essai1_0009.edf",
            "essai1_0020.edf",
            "essai1_0031.edf",
            "essai1_0042.edf",
            "essai1_0053.edf",
            "essai1_0064.edf",
            "essai1_0075.edf",
            "essai1_0086.edf",
            "essai1_0097.edf",
            "essai1_.rec",
            "essai1_0000.edf",
            "essai1_0010.edf",
            "essai1_0021.edf",
            "essai1_0032.edf",
            "essai1_0043.edf",
            "essai1_0054.edf",
            "essai1_0065.edf",
            "essai1_0076.edf",
            "essai1_0087.edf",
            "essai1_0098.edf",
            "essai1_slice_1023.edf",
            "essai1_0001.edf",
            "essai1_0011.edf",
            "essai1_0022.edf",
            "essai1_0033.edf",
            "essai1_0044.edf",
            "essai1_0055.edf",
            "essai1_0066.edf",
            "essai1_0077.edf",
            "essai1_0088.edf",
            "essai1_0099.edf",
            "essai1_slice.info",
            "essai1_0001.par",
            "essai1_0012.edf",
            "essai1_0023.edf",
            "essai1_0034.edf",
            "essai1_0045.edf",
            "essai1_0056.edf",
            "essai1_0067.edf",
            "essai1_0078.edf",
            "essai1_0089.edf",
            "essai1_0100.edf",
            "essai1_slice.par",
            "essai1_0002.edf",
            "essai1_0013.edf",
            "essai1_0024.edf",
            "essai1_0035.edf",
            "essai1_0046.edf",
            "essai1_0057.edf",
            "essai1_0068.edf",
            "essai1_0079.edf",
            "essai1_0090.edf",
            "essai1_0101.edf",
            "essai1_slice.xml",
            "essai1_0003.edf",
            "essai1_0014.edf",
            "essai1_0025.edf",
            "essai1_0036.edf",
            "essai1_0047.edf",
            "essai1_0058.edf",
            "essai1_0069.edf",
            "essai1_0080.edf",
            "essai1_0091.edf",
            "essai1_0102.edf",
            "essai1_.xml",
            "essai1_0004.edf",
            "essai1_0015.edf",
            "essai1_0026.edf",
            "essai1_0037.edf",
            "essai1_0048.edf",
            "essai1_0059.edf",
            "essai1_0070.edf",
            "essai1_0081.edf",
            "essai1_0092.edf",
            "essai1_0103.edf",
            "histogram_essai1_slice",
            "essai1_0005.edf",
            "essai1_0016.edf",
            "essai1_0027.edf",
            "essai1_0038.edf",
            "essai1_0049.edf",
            "essai1_0060.edf",
            "essai1_0071.edf",
            "essai1_0082.edf",
            "essai1_0093.edf",
            "essai1_0104.edf",
            "machinefile",
            "essai1_0006.edf",
            "essai1_0017.edf",
            "essai1_0028.edf",
            "essai1_0039.edf",
            "essai1_0050.edf",
            "essai1_0061.edf",
            "essai1_0072.edf",
            "essai1_0083.edf",
            "essai1_0094.edf",
            "essai1_.cfg",
            "pyhst_out.txt",
            "essai1_0007.edf",
            "essai1_0018.edf",
            "essai1_0029.edf",
            "essai1_0040.edf",
            "essai1_0051.edf",
            "essai1_0062.edf",
            "essai1_0073.edf",
            "essai1_0084.edf",
            "essai1_0095.edf",
        ]

        nbRadio = 0
        for f in files:
            nbRadio += EDFTomoScan.is_a_proj_path(f, "essai1_")
        self.assertTrue(nbRadio == 105)


class TestGetReconstructionPath(unittest.TestCase):
    """Test static method getReconstructionPaths for EDFTomoScan"""

    def setUp(self):
        self.tmpDir = tempfile.mkdtemp()
        self.dir = os.path.join(self.tmpDir, "scan_3")
        os.makedirs(self.dir)

    def tearDown(self):
        shutil.rmtree(self.tmpDir)

    def test1(self):
        """Make sure he can read 3 simple slice"""
        file_desc = fabio.edfimage.EdfImage(data=numpy.random.random((100, 100)))

        file_desc.write(os.path.join(self.dir, "scan_3slice_0050.edf"))
        file_desc.write(os.path.join(self.dir, "scan_3slice_0000.edf"))
        file_desc.write(os.path.join(self.dir, "scan_3slice_0010.edf"))
        assert len(os.listdir(self.dir)) == 3
        self.assertEqual(
            len(
                EDFTomoScan.get_reconstructions_paths(
                    os.path.join(self.tmpDir, "scan_3")
                )
            ),
            3,
        )

    def test2(self):
        """Make sure he can read a paganin and non-paganin at the same time"""
        file_desc = fabio.edfimage.EdfImage(data=numpy.random.random((100, 100)))

        file_desc.write(os.path.join(self.dir, "scan_3slice_0050.edf"))
        file_desc.write(os.path.join(self.dir, "scan_3slice_pag_0050.edf"))
        assert len(os.listdir(self.dir)) == 2
        self.assertEqual(
            len(
                EDFTomoScan.get_reconstructions_paths(
                    os.path.join(self.tmpDir, "scan_3")
                )
            ),
            2,
        )


class TestPaganinPath(unittest.TestCase):
    """Test static method getIndexReconstructed from EDFTomoScan"""

    def setUp(self):
        self.tmpDir = tempfile.mkdtemp()

        self.dir = os.path.join(self.tmpDir, "scan25")
        os.makedirs(self.dir)

    def tearDown(self):
        shutil.rmtree(self.tmpDir)

    def testPagFile(self):
        open(os.path.join(self.dir, "scan25slice_pag_db0500_0115.edf"), "a").close()
        self.assertEqual(len(EDFTomoScan.get_reconstructions_paths(self.dir)), 1)
        self.assertEqual(
            EDFTomoScan.get_index_reconstructed(
                "scan25slice_pag_db0500_0115.edf", scanID=self.dir
            ),
            115,
        )


def test_get_sinogram(tmp_path):
    edf_scan_folder = tmp_path / "scan_dir"
    os.makedirs(edf_scan_folder, exist_ok=True)

    mock = MockEDF(
        scan_path=edf_scan_folder, scene="arange", n_radio=4, n_ini_radio=4, dim=2
    )
    mock.end_acquisition()
    scan = EDFTomoScan(scan=mock.scan_path)
    assert len(scan.projections) == 4
    assert scan.tomo_n == 4

    sinogram_1 = scan.get_sinogram(line=1)
    assert sinogram_1.shape == (4, 2)
    numpy.testing.assert_array_equal(
        sinogram_1,
        numpy.array([[2, 3], [6, 7], [10, 11], [14, 15]]),
    )
